#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File: utils.py
#
# Copyright 2020 Costas Tyfoxylos
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
#

"""
Main code for utils.

.. _Google Python Style Guide:
   http://google.github.io/styleguide/pyguide.html

"""

import json
import logging

__author__ = '''Costas Tyfoxylos <ctyfoxylos@schubergphilis.com>'''
__docformat__ = '''google'''
__date__ = '''24-12-2020'''
__copyright__ = '''Copyright 2020, Costas Tyfoxylos'''
__credits__ = ["Costas Tyfoxylos"]
__license__ = '''MIT'''
__maintainer__ = '''Costas Tyfoxylos'''
__email__ = '''<ctyfoxylos@schubergphilis.com>'''
__status__ = '''Development'''  # "Prototype", "Development", "Production".

# This is the main prefix used for logging
LOGGER_BASENAME = '''utils'''
LOGGER = logging.getLogger(LOGGER_BASENAME)
LOGGER.addHandler(logging.NullHandler())


class HarParser:
    """Parses a provided har file."""

    def __init__(self, har_file):
        self.data = self._parse(har_file)

    @staticmethod
    def _parse(har_file):
        try:
            data = json.load(open(har_file, 'r'))
        except Exception:
            raise ValueError(f'Could not read or parse file: {har_file}')
        return data

    def _get_service_calls(self, service):
        calls = [entry for entry in self.data['log']['entries']
                 if any(['aws.amazon.com/oauth' in entry['request']['url'],
                         all([f'aws.amazon.com/{service}' in entry['request']['url'],
                              f'aws.amazon.com/{service}/api' not in entry['request']['url']])])]
        return [] if not any([True for entry in calls if service in entry["request"]["url"]]) else calls

    @staticmethod
    def _get_text_from_calls(calls):
        text = ''
        for entry in calls:
            text += f'URL : {entry["request"]["url"]}\n'
            text += '\tRequest Headers :\n'
            for header in entry['request']['headers']:
                text += f'\t\t{header["name"]} : {header["value"]}\n'
            text += "\tRequest Cookies :\n"
            for cookie in entry['request']['cookies']:
                text += f'\t\t{cookie["name"]} : {cookie["value"]}\n'
            text += "\tResponse Headers :\n"
            for header in entry['response']['headers']:
                text += f'\t\t{header["name"]} : {header["value"]}\n'
            text += "\tResponse Cookies :\n"
            for cookie in entry['response']['cookies']:
                text += f'\t\t{cookie["name"]} : {cookie["value"]}\n'
        return text

    def get_communication_for_console(self):
        """Returns a text of the communication of a valid login to console.

        Returns:
            text (str): Returns a text of the communication of a valid login to console.

        """
        return self._get_text_from_calls(self._get_service_calls('console'))

    def get_communication_for_control_tower(self):
        """Returns a text of the communication of a valid login to control tower.

        Returns:
            text (str): Returns a text of the communication of a valid login to control tower.

        """
        return self._get_text_from_calls(self._get_service_calls('controltower'))

    def get_communication_for_sso(self):
        """Returns a text of the communication of a valid login to single sign on.

        Returns:
            text (str): Returns a text of the communication of a valid login to single sign on.

        """
        return self._get_text_from_calls(self._get_service_calls('singlesignon'))

    def get_communication_for_billing(self):
        """Returns a text of the communication of a valid login to billing.

        Returns:
            text (str): Returns a text of the communication of a valid login to billing.

        """
        return self._get_text_from_calls(self._get_service_calls('billing'))

    def render_communication_for_console(self):
        """Prints a text of the communication of a valid login to console.

        Returns:
            None

        """
        print(self.get_communication_for_console())

    def render_communication_for_control_tower(self):
        """Prints a text of the communication of a valid login to control tower.

        Returns:
            None

        """
        print(self.get_communication_for_control_tower())

    def render_communication_for_sso(self):
        """Prints a text of the communication of a valid login to single sign on.

        Returns:
            None

        """
        print(self.get_communication_for_sso())

    def render_communication_for_billing(self):
        """Prints a text of the communication of a valid login to billing.

        Returns:
            None

        """
        print(self.get_communication_for_billing())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File: __init__.py
#
# Copyright 2020 Costas Tyfoxylos
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
#

"""
controltower module.

Import all parts from controltower here

.. _Google Python Style Guide:
   http://google.github.io/styleguide/pyguide.html
"""
from .controltowerexceptions import (UnsupportedTarget,
                                     OUCreating,
                                     NoServiceCatalogAccess,
                                     NonExistentSCP,
                                     NoSuspendedOU,
                                     ServiceCallFailed,
                                     ControlTowerBusy,
                                     ControlTowerNotDeployed,
                                     PreDeployValidationFailed,
                                     EmailCheckFailed,
                                     EmailInUse,
                                     UnavailableRegion,
                                     RoleCreationFailure)
from .controltower import ControlTower

__author__ = '''Costas Tyfoxylos <ctyfoxylos@schubergphilis.com>'''
__docformat__ = '''google'''
__date__ = '''18-02-2020'''
__copyright__ = '''Copyright 2020, Costas Tyfoxylos'''
__license__ = '''MIT'''
__maintainer__ = '''Costas Tyfoxylos'''
__email__ = '''<ctyfoxylos@schubergphilis.com>'''
__status__ = '''Development'''  # "Prototype", "Development", "Production".

# This is to 'use' the module(s), so lint doesn't complain
assert ControlTower

assert UnsupportedTarget
assert OUCreating
assert NoServiceCatalogAccess
assert NonExistentSCP
assert NoSuspendedOU
assert ServiceCallFailed
assert ControlTowerBusy
assert ControlTowerNotDeployed
assert PreDeployValidationFailed
assert EmailCheckFailed
assert EmailInUse
assert UnavailableRegion
assert RoleCreationFailure



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File: billing.py
#
# Copyright 2021 Costas Tyfoxylos
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
#

"""
Main code for billing.

.. _Google Python Style Guide:
   http://google.github.io/styleguide/pyguide.html

"""

import logging
import time
from datetime import datetime, timedelta

from bs4 import BeautifulSoup as Bfs

from awsapilib.authentication import Authenticator, LoggerMixin
from awsapilib.authentication import InvalidCredentials

from .billingexceptions import (InvalidCountryCode,
                                NonEditableSetting,
                                IAMAccessDenied,
                                InvalidCurrency,
                                ServerError)


__author__ = '''Costas Tyfoxylos <ctyfoxylos@schubergphilis.com>'''
__docformat__ = '''google'''
__date__ = '''30-03-2021'''
__copyright__ = '''Copyright 2021, Costas Tyfoxylos'''
__credits__ = ["Costas Tyfoxylos"]
__license__ = '''MIT'''
__maintainer__ = '''Costas Tyfoxylos'''
__email__ = '''<ctyfoxylos@schubergphilis.com>'''
__status__ = '''Development'''  # "Prototype", "Development", "Production".

# This is the main prefix used for logging
LOGGER_BASENAME = '''billing'''
LOGGER = logging.getLogger(LOGGER_BASENAME)
LOGGER.addHandler(logging.NullHandler())


class Tax(LoggerMixin):
    """Models the tax settings of the billing console."""

    def __init__(self, billing):
        self._billing = billing
        self._endpoint = f'{self._billing.rest_api}/taxexemption/heritage'
        self._available_country_codes = None

    @property
    def available_country_codes_eu(self):
        """The available country codes of the tax settings for eu.

        Returns:
            codes (list): Available country codes

        """
        if self._available_country_codes is None:
            url = f'{self._billing.rest_api}/taxexemption/eu/vat/countries'
            response = self._billing.session.get(url)
            if response.status_code == 401:
                raise InvalidCredentials
            if not response.ok:
                self.logger.error(f'Failed to retrieve inheritance state, response: {response.text}')
                return None
            self._available_country_codes = response.json().get('supportedCountryCodes', [])
        return self._available_country_codes

    @property
    def inheritance(self):
        """The inheritance settings of the billing preferences.

        Returns:
            setting (bool): True if set, False otherwise.

        """
        response = self._billing.session.get(self._endpoint)
        if response.status_code == 401:
            raise InvalidCredentials
        if not response.ok:
            self.logger.error(f'Failed to retrieve inheritance state, response: {response.text}')
            return None
        return response.json().get('customerHeritageStatus', '') == 'OptIn'

    @inheritance.setter
    def inheritance(self, value: bool):
        """The inheritance settings setter of the billing preferences.

        Returns:
            None

        """
        if self.inheritance == value:
            return
        self._is_editable()
        parameters = {'heritageStatus': 'OptIn' if value else 'OptOut'}
        response = self._billing.session.post(self._endpoint, params=parameters)
        if not response.ok:
            self.logger.error(f'Failed to retrieve inheritance state, response: {response.text}')

    def _is_editable(self):
        parameters = {'heritageStatus': 'OptIn'}
        response = self._billing.session.get(self._endpoint, params=parameters)
        if not response.json().get('heritageStatusEditable'):
            timestamp = response.json().get('effectiveTimestamp')
            unlock_time = (datetime.fromtimestamp(timestamp / 1000) + timedelta(minutes=15))
            wait_time = unlock_time - datetime.now()
            raise NonEditableSetting(f'API is not enabled for {wait_time} more.')
        return True

    #  pylint: disable=too-many-arguments
    def set_information(self, address, city, postal_code, legal_name, vat_number, country_code, state=None):
        """The inheritance settings setter of the billing preferences.

        Returns:
            None

        """
        country_code = country_code.upper()
        if country_code not in self.available_country_codes_eu:
            raise InvalidCountryCode(f'{country_code} provided is not valid. '
                                     f'Valid ones are {self.available_country_codes_eu}')
        payload = {'address': {'addressLine1': address,
                               'addressLine2': None,
                               'city': city,
                               'countryCode': country_code,
                               'postalCode': postal_code,
                               'state': state,
                               },
                   'authority': {'country': country_code,
                                 'state': None},
                   'legalName': legal_name,
                   'localTaxRegistration': False,
                   'registrationId': vat_number,
                   }
        url = f'{self._billing.rest_api}/taxexemption/eu/vat/information'
        response = self._billing.session.put(url, json=payload)
        if not response.ok:
            self.logger.error(f'Failed to set information, response: {response.text}')
        return response.ok


class Preferences(LoggerMixin):
    """Models the preferences of the billing console."""

    def __init__(self, billing):
        self._billing = billing

    @property
    def _preferences_endpoint(self):
        return f'{self._billing.rest_api}/preferences'

    @property
    def _values(self):
        response = self._billing.session.get(self._preferences_endpoint)
        if not response.ok:
            self.logger.error(f'Failed to retrieve inheritance state, response: {response.text}')
            return {}
        return response.json()

    @property
    def pdf_invoice_by_mail(self):
        """The setting of the pdf invoice by email.

        Returns:
            setting (bool): True if set, False otherwise.

        """
        values = self._values
        if not values:
            raise SystemError('Could not retrieve the preferences!')
        return values.get('pdfInvoiceByEmail') == 'Y'

    @pdf_invoice_by_mail.setter
    def pdf_invoice_by_mail(self, value: bool):
        """The setting for the setting of the pdf invoice by email.

        Returns:
            None.

        """
        if self.pdf_invoice_by_mail == value:
            return
        payload = {'pdfInvoiceByEmail': 'Y' if value else 'N'}
        response = self._billing.session.put(self._preferences_endpoint, json=payload)
        if not response.ok:
            self.logger.error(f'Failed to retrieve inheritance state, response: {response.text}')

    @property
    def credit_sharing(self):
        """The setting of the credit sharing.

        Returns:
            setting (bool): True if set, False otherwise.

        """
        endpoint = f'{self._billing.rest_api}/sharingpreferences/getcreditsharing'
        response = self._billing.session.get(endpoint)
        if not response.ok:
            self.logger.error(f'Failed to retrieve credit sharing state, response: {response.text}')
            return {}
        return response.json().get('creditEnabled')

    @credit_sharing.setter
    def credit_sharing(self, value: bool):
        """The setter of the setting of the credit sharing.

        Returns:
            None.

        """
        if self.credit_sharing == value:
            return
        endpoint = f'{self._billing.rest_api}/sharingpreferences/setcreditsharing'
        payload = {'creditEnabled': bool(value)}
        response = self._billing.session.put(endpoint, json=payload)
        if not response.ok:
            self.logger.error(f'Failed to retrieve credit sharing state, response: {response.text}')


class Billing(LoggerMixin):
    """Models Control Tower by wrapping around service catalog."""

    def __init__(self, arn, region=None):
        self.aws_authenticator = Authenticator(arn)
        self.session = self._get_authenticated_session()
        self.region = region or self.aws_authenticator.region
        self.rest_api = 'https://console.aws.amazon.com/billing/rest/v1.0'
        self._sor_info_ = None
        self._payment_instrument_ids = None
        self._marketplace_id = None

    def _get_authenticated_session(self):
        return self.aws_authenticator.get_billing_authenticated_session()

    @property
    def account_id(self):
        """Account id."""
        return self._sor_info.get('accountId')

    @property
    def sor_id(self):
        """Sor id."""
        return self._sor_info.get('sor', {}).get('sorId')

    @property
    def _sor_info(self):
        if self._sor_info_ is None:
            url = f'{self.rest_api}/sellerofrecord/getsorbyaccount'
            response = self.session.get(url)
            if response.status_code == 401:
                raise InvalidCredentials
            if not response.ok:
                self.logger.error(f'Could not retrieve sor info, response: {response.text}')
                self._sor_info_ = []
            self._sor_info_ = response.json()
        return self._sor_info_

    @property
    def tax(self):
        """Tax settings.

        Returns:
            tax (Tax): The tax settings object.

        """
        return Tax(self)

    @property
    def preferences(self):
        """Preferences settings.

        Returns:
            preferences (Preferences): The preferences settings object.

        """
        return Preferences(self)

    @property
    def currency(self):
        """Currency settings.

        Returns:
            currency (str): The currency set.

        """
        url = f'{self.rest_api}/account/fxpaymentinfopapyrus'
        response = self.session.get(url)
        if not response.ok:
            self.logger.error(f'Failed to retrieve currency setting, response: {response.text}')
            return None
        return response.json().get('currencyPreference')

    @currency.setter
    def currency(self, value):
        """Setter for currency settings.

        Returns:
            None

        """
        url = f'{self.rest_api}/account/currencypreference'
        response = self.session.put(url, json=value.upper())
        if not response.ok:
            if response.json().get('type') == 'InvalidParameterException':
                raise InvalidCurrency(value)
            self.logger.error(f'Failed to set currency setting, response: {response.text}')

    def _validate_iam_access(self):
        url = f'{self.rest_api}/account/iamaccess'
        response = self.session.get(url)
        if not response.ok:
            self.logger.error(f'Failed to get iam access settings, response: {response.text}')
            return {}
        if response.json().get('type') == 'AccessDeniedException':
            raise IAMAccessDenied
        return response.json()

    @property
    def iam_access(self):
        """IAM access to billing setting."""
        return self._validate_iam_access().get('billingConsoleAccessEnabled', False)

    @iam_access.setter
    def iam_access(self, value):
        """IAM access to billing setting."""
        data = self._validate_iam_access()
        if data:
            url = f'{self.rest_api}/account/iamaccess'
            data.update({'billingConsoleAccessEnabled': bool(value)})
            response = self.session.put(url, json=data)
            if not response.ok:
                self.logger.error(f'No IAM role access provided to the console, response: {response.text}')

    @property
    def _region_states(self):
        url = f'{self.rest_api}/account/accountregionstates'
        response = self.session.get(url)
        if response.status_code == 401:
            raise InvalidCredentials
        if not response.ok:
            self.logger.error(f'Could not retrieve region states, response: {response.text}')
            return []
        return response.json().get('accountRegionStateList', [])

    @property
    def enabled_region_states(self):
        """Enabled region states."""
        return [region.get('regionName') for region in self._region_states
                if region.get('regionState') == 'ENABLED']

    @property
    def disabled_region_states(self):
        """Disabled region states."""
        return [region.get('regionName') for region in self._region_states
                if region.get('regionState') == 'DISABLED']

    @property
    def payment_cards(self):
        """Payment cards."""
        if self._payment_instrument_ids is None:
            url = 'https://console.aws.amazon.com/billing/rest/ppg-proxy'
            headers = {'x-requested-with': 'XMLHttpRequest',
                       'Operation': 'AWSPaymentPreferenceGateway.Get'}
            payload = {'content': {'Input': {'arn': f'arn:aws:payments:us-east-1:{self.account_id}:'
                                                    f'paymentpreference:PaymentInstrument'},
                                   'Operation': 'com.amazon.aws.payments.gateway.coral.'
                                                'paymentpreference.operations#Get',
                                   'Service': 'com.amazon.aws.payments.gateway.coral.paymentpreference.'
                                              'service#AWSPaymentPreferenceGateway'},
                       'headers': {'Content-Type': 'application/json',
                                   'X-Amz-Date': time.strftime("%a, %d %b %Y %I:%M:%S %Z", time.gmtime()),
                                   'X-Amz-Target': 'AWSPaymentPreferenceGateway.Get'},
                       'region': 'us-east-1'}
            response = self.session.post(url, headers=headers, json=payload)
            if response.status_code == 401:
                raise InvalidCredentials
            if not response.ok:
                self.logger.error(f'Could not retrieve payment instrument id, response: {response.text}')
                raise ServerError
            metadata = response.json().get('Output', {}).get('paymentPreferenceWithMetadata', {})
            self._payment_instrument_ids = [PaymentCard(self, data)
                                            for data in metadata.get('value', {}).get('chargeInstruments', [])]
        return self._payment_instrument_ids

    @property
    def market_place_id(self):
        """Marker place id of account."""
        if self._marketplace_id is None:
            url = 'https://console.aws.amazon.com/billing/home?'
            response = self.session.get(url)
            if response.status_code == 401:
                raise InvalidCredentials
            if not response.ok:
                self.logger.error(f'Could not retrieve market place id, response: {response.text}')
                raise ServerError
            soup = Bfs(response.text, features="html.parser")
            self._marketplace_id = soup.find('input', {'id': 'marketPlace'}).attrs.get('value')
        return self._marketplace_id

    # def get_attribute(self, path):
    #     response = self.session.get(f'{self.rest_api}/{path}')
    #     return response


class PaymentCard(LoggerMixin):
    """Models a payment card."""

    def __init__(self, billing, data):
        self._billing = billing
        self._arn = data.get('arn')
        self._data_ = None

    @property
    def _data(self):
        if self._data_ is None:
            url = f'{self._billing.rest_api}/billingcontactaddress/get'
            parameters = {'marketplaceId': self._billing.market_place_id,
                          'piArn': self._arn}
            response = self._billing.session.get(url, params=parameters)
            if response.status_code == 401:
                raise InvalidCredentials
            if not response.ok:
                self.logger.error(f'Could not retrieve market place id, response: {response.text}')
                raise ServerError
            self._data_ = response.json()
        return self._data_

    @property
    def _address(self):
        return self._data.get('address', {})

    @property
    def address_id(self):
        """Address id."""
        return self._address.get('addressId')

    @property
    def address_line_1(self):
        """First line of the address settings."""
        return self._address.get('addressLine1')

    @property
    def address_line_2(self):
        """Second line of the address settings."""
        return self._address.get('addressLine2')

    @property
    def city(self):
        """City."""
        return self._address.get('city')

    @property
    def company(self):
        """Company."""
        return self._address.get('company')

    @property
    def country_code(self):
        """Country code."""
        return self._address.get('countryCode')

    @property
    def email_address_list(self):
        """Email address list."""
        return self._address.get('emailAddressList', [])

    @property
    def full_name(self):
        """Full name."""
        return self._address.get('fullName')

    @property
    def phone_number(self):
        """Phone number."""
        return self._address.get('phoneNumber')

    @property
    def postal_code(self):
        """Postal code."""
        return self._address.get('postalCode')

    @property
    def state(self):
        """State."""
        return self._address.get('state')

    @property
    def payment_instrument_arn(self):
        """Payment instrument arn."""
        return self._data.get('paymentInstrumentArn')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File: __init__.py
#
# Copyright 2020 Costas Tyfoxylos
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
#

"""
resources module.

Import all parts from resources here

.. _Google Python Style Guide:
   http://google.github.io/styleguide/pyguide.html
"""

from time import sleep

from awsapilib.authentication import LoggerMixin
from awsapilib.controltower.controltowerexceptions import (NonExistentSCP,
                                                           ControlTowerBusy,
                                                           NoSuspendedOU)

__author__ = '''Costas Tyfoxylos <ctyfoxylos@schubergphilis.com>'''
__docformat__ = '''google'''
__date__ = '''18-02-2020'''
__copyright__ = '''Copyright 2020, Costas Tyfoxylos'''
__license__ = '''MIT'''
__maintainer__ = '''Costas Tyfoxylos'''
__email__ = '''<ctyfoxylos@schubergphilis.com>'''
__status__ = '''Development'''  # "Prototype", "Development", "Production".


class AccountFactory:  # pylint: disable=too-few-public-methods, too-many-instance-attributes
    """Models the account factory data of service catalog."""

    def __init__(self, service_catalog_client, data):
        self._service_catalog = service_catalog_client
        self._data = data
        self.has_default_path = self._data.get('HasDefaultPath')
        self.id = self._data.get('Id')  # pylint: disable=invalid-name
        self.name = self._data.get('Name')
        self.owner = self._data.get('Owner')
        self.product_id = self._data.get('ProductId')
        self.short_description = self._data.get('ShortDescription')
        self.type = self._data.get('Type')


class ServiceControlPolicy:
    """Models the account factory data of service catalog."""

    def __init__(self, data):
        self._data = data

    @property
    def arn(self):
        """Arn."""
        return self._data.get('Arn')

    @property
    def aws_managed(self):
        """Aws Managed."""
        return self._data.get('AwsManaged')

    @property
    def description(self):
        """Description."""
        return self._data.get('Description')

    @property
    def id(self):  # pylint: disable=invalid-name
        """Id."""
        return self._data.get('Id')

    @property
    def name(self):
        """Name."""
        return self._data.get('Name')

    @property
    def type(self):
        """Type."""
        return self._data.get('Type')


class GuardRail(LoggerMixin):
    """Models the guard rail data."""

    def __init__(self, control_tower, data):
        self.control_tower = control_tower
        self._data_ = data

    @property
    def _data(self):
        """The data of the guard rail as returned by the api."""
        return self._data_

    @property
    def behavior(self):
        """Behavior."""
        return self._data_.get('Behavior')

    @property
    def category(self):
        """Category."""
        return self._data_.get('Category')

    @property
    def description(self):
        """Description."""
        return self._data_.get('Description')

    @property
    def display_name(self):
        """DisplayName."""
        return self._data_.get('DisplayName')

    @property
    def name(self):
        """Name."""
        return self._data_.get('Name')

    @property
    def provider(self):
        """Provider."""
        return self._data_.get('Provider')

    @property
    def regional_preference(self):
        """Regional preference."""
        return self._data_.get('RegionalPreference')

    @property
    def type(self):
        """Type."""
        return self._data_.get('Type')

    @property
    def compliancy_status(self):
        """Compliancy status."""
        payload = self.control_tower._get_api_payload(content_string={'GuardrailName': self.name},  # pylint: disable=protected-access
                                                      target='getGuardrailComplianceStatus')
        self.logger.debug('Trying to get the compliancy status with payload "%s"', payload)
        response = self.control_tower.session.post(self.control_tower.url, json=payload)
        if not response.ok:
            self.logger.error('Failed to get the drift message of the landing zone with response status "%s" and '
                              'response text "%s"',
                              response.status_code, response.text)
            return None
        return response.json().get('ComplianceStatus')


class CoreAccount:
    """Models the core landing zone account data."""

    def __init__(self, control_tower, account_label, data):
        self.control_tower = control_tower
        self._label = account_label
        self._data_ = data

    @property
    def _data(self):
        """The data of the account as returned by the api."""
        return self._data_

    @property
    def label(self):
        """Account label."""
        return self._label

    @property
    def email(self):
        """Email."""
        return self._data_.get('AccountEmail')

    @property
    def id(self):  # pylint: disable=invalid-name
        """Id."""
        return self._data_.get('AccountId')

    @property
    def core_resource_mappings(self):
        """Core resource mappings."""
        return self._data_.get('CoreResourceMappings')

    @property
    def stack_set_arn(self):
        """Stack set arn."""
        return self._data_.get('StackSetARN')


class ControlTowerAccount(LoggerMixin):  # pylint: disable=too-many-public-methods
    """Models the account data."""

    def __init__(self, control_tower, data, info_polling_interval=30):
        self.control_tower = control_tower
        self.service_catalog = control_tower.service_catalog
        self.organizations = control_tower.organizations
        self._data_ = data
        self._service_catalog_data_ = None
        self._record_data_ = None
        self._info_polling_interval = info_polling_interval

    @property
    def _data(self):
        """The data of the account as returned by the api."""
        return self._data_

    @property
    def _service_catalog_data(self):
        if self._service_catalog_data_ is None:
            data = self.service_catalog.search_provisioned_products(Filters={'SearchQuery': [f'physicalId:{self.id}']})
            if not data.get('TotalResultsCount'):
                self._service_catalog_data_ = {}
            else:
                self._service_catalog_data_ = data.get('ProvisionedProducts', [{}]).pop()
        return self._service_catalog_data_

    @property
    def _record_data(self):
        if self._record_data_ is None:
            if not self.last_record_id:
                self._record_data_ = {}
            else:
                self._record_data_ = self.service_catalog.describe_record(Id=self.last_record_id)
        return self._record_data_

    @property
    def email(self):
        """Email."""
        return self._data_.get('AccountEmail')

    @property
    def id(self):  # pylint: disable=invalid-name
        """Id."""
        return self._data_.get('AccountId')

    @property
    def name(self):
        """Name."""
        return self._data_.get('AccountName')

    @property
    def arn(self):
        """Arn."""
        return self._data_.get('Arn')

    @property
    def owner(self):
        """Owner."""
        return self._data_.get('Owner')

    @property
    def provision_state(self):
        """Provision state."""
        return self._data_.get('ProvisionState')

    @property
    def status(self):
        """Status."""
        return self._data_.get('Status')

    @property
    def landing_zone_version(self):
        """Landing zone version."""
        return self._data_.get('DeployedLandingZoneVersion')

    @property
    def has_available_update(self):
        """If the account is behind the landing zone version."""
        if self.provision_state == 'PROVISION_FAILED':
            return False
        return float(self.landing_zone_version) < float(self.control_tower.landing_zone_version)

    @property
    def guardrail_compliance_status(self):
        """Retrieves the guardrail compliancy status for the account.

        Returns:
            status (str): COMPLIANT|NON COMPLIANT

        """
        payload = self.control_tower._get_api_payload(content_string={'AccountId': self.id},  # pylint: disable=protected-access
                                                      target='getGuardrailComplianceStatus')
        response = self.control_tower.session.post(self.control_tower.url, json=payload)
        if not response.ok:
            self.logger.error('Failed to get compliancy status from api.')
            return False
        return response.json().get('ComplianceStatus')

    @property
    def organizational_unit(self):
        """Organizational Unit."""
        return self.control_tower.get_organizational_unit_by_id(self._data_.get('ParentOrganizationalUnitId'))

    @property
    def stack_arn(self):
        """Stack Arn."""
        return self._service_catalog_data.get('Arn')

    @property
    def created_time(self):
        """Created Time."""
        return self._service_catalog_data.get('CreatedTime')

    @property
    def service_catalog_id(self):
        """Service Catalog ID."""
        return self._service_catalog_data.get('Id')

    @property
    def idempotency_token(self):
        """Idempotency Token."""
        return self._service_catalog_data.get('IdempotencyToken')

    @property
    def last_record_id(self):
        """Last Record ID."""
        return self._service_catalog_data.get('LastRecordId')

    @property
    def physical_id(self):
        """Physical ID."""
        return self._service_catalog_data.get('PhysicalId')

    @property
    def service_catalog_product_id(self):
        """Service catalog product ID."""
        return self._service_catalog_data.get('ProductId')

    @property
    def provisioning_artifact_id(self):
        """Provisioning artifact ID."""
        return self._service_catalog_data.get('ProvisioningArtifactId')

    @property
    def service_catalog_tags(self):
        """Service catalog tags."""
        return self._service_catalog_data.get('Tags')

    @property
    def service_catalog_type(self):
        """Service catalog type."""
        return self._service_catalog_data.get('Type')

    @property
    def service_catalog_status(self):
        """Service catalog status."""
        return self._service_catalog_data.get('Status')

    @property
    def service_catalog_user_arn(self):
        """Service catalog user arn."""
        return self._service_catalog_data.get('UserArn')

    @property
    def user_arn_session(self):
        """User arn session."""
        return self._service_catalog_data.get('UserArnSession')

    def _refresh(self):
        self._data_ = self.control_tower.get_account_by_id(self.id)._data  # pylint: disable=protected-access
        self._record_data_ = None
        self._service_catalog_data_ = None

    def _get_record_entry(self, output_key):
        return next((entry for entry in self._record_data.get('RecordOutputs', [])
                     if entry.get('OutputKey', '') == output_key), {})

    @property
    def sso_user_email(self):
        """SSO user email."""
        return self._get_record_entry(output_key='SSOUserEmail').get('OutputValue')

    @property
    def sso_user_portal(self):
        """SSO user portal."""
        return self._get_record_entry(output_key='SSOUserPortal').get('OutputValue')

    def detach_service_control_policy(self, name):
        """Detaches a Service Control Policy from the account.

        Args:
            name (str): The name of the SCP to detach

        Returns:
            result (bool): True on success, False otherwise.

        """
        return self._action_service_control_policy('detach', name)

    def attach_service_control_policy(self, name):
        """Attaches a Service Control Policy to the account.

        Args:
            name (str): The name of the SCP to attach

        Returns:
            result (bool): True on success, False otherwise.

        """
        return self._action_service_control_policy('attach', name)

    def _action_service_control_policy(self, action, scp_name):
        scp = self.control_tower.get_service_control_policy_by_name(scp_name)
        if not scp:
            raise NonExistentSCP(scp_name)
        response = getattr(self.organizations, f'{action}_policy')(PolicyId=scp.id, TargetId=self.id)
        if not response.get('ResponseMetadata', {}).get('HTTPStatusCode') == 200:
            self.logger.error('Failed to %s SCP "%s" to account with response "%s"', action, scp.name, response)
            return False
        self.logger.debug('Successfully %sed SCP "%s" to account', action, scp.name)
        return True

    def _terminate(self):
        """Terminates an account that is in error.

        Returns:
            response (dict): The response from the api of the termination request.

        """
        return self.service_catalog.terminate_provisioned_product(ProvisionedProductId=self.service_catalog_id)

    def delete(self, suspended_ou_name=None):
        """Delete."""
        if not suspended_ou_name:
            return self._terminate()
        suspended_ou = self.control_tower.get_organizational_unit_by_name(suspended_ou_name)
        if not suspended_ou:
            raise NoSuspendedOU(suspended_ou_name)
        self._terminate()
        while self.control_tower.busy:
            self.logger.debug('Waiting for control tower to terminate the account...')
            sleep(self._info_polling_interval)
        self.logger.debug('Moving account from root OU to %s', suspended_ou_name)
        self.organizations.move_account(AccountId=self.id,
                                        SourceParentId=self.control_tower.root_ou.id,
                                        DestinationParentId=suspended_ou.id)
        self.logger.debug('Attaching SCP %s to account', suspended_ou_name)
        self.attach_service_control_policy(suspended_ou_name)
        self.logger.debug('Detaching full access SCP from account')
        self.detach_service_control_policy('FullAWSAccess')
        return True

    def update(self):
        """Updates the account in service catalog.

        Returns:
            True if the call succeeded False otherwise

        """
        if not self.has_available_update:
            return True
        if self.control_tower.busy:
            raise ControlTowerBusy
        arguments = {'ProductId': self.control_tower._account_factory.product_id,  # pylint: disable=protected-access
                     'ProvisionedProductName': self.name,
                     'ProvisioningArtifactId': self.control_tower._active_artifact.get('Id'),  # pylint: disable=protected-access
                     'ProvisioningParameters': [{'Key': 'AccountName',
                                                 'Value': self.name,
                                                 'UsePreviousValue': True},
                                                {'Key': 'AccountEmail',
                                                 'Value': self.email,
                                                 'UsePreviousValue': True},
                                                {'Key': 'SSOUserFirstName',
                                                 'Value': 'Control',
                                                 'UsePreviousValue': True},
                                                {'Key': 'SSOUserLastName',
                                                 'Value': 'Tower',
                                                 'UsePreviousValue': True},
                                                {'Key': 'SSOUserEmail',
                                                 'Value': self.email,
                                                 'UsePreviousValue': True},
                                                {'Key': 'ManagedOrganizationalUnit',
                                                 'Value': self.organizational_unit.name,
                                                 'UsePreviousValue': True}]}
        response = self.service_catalog.update_provisioned_product(**arguments)
        return response.get('ResponseMetadata', {}).get('HTTPStatusCode') == 200


class OrganizationsOU:
    """Model the data of an Organizations managed OU."""

    def __init__(self, data):
        self._data = data

    @property
    def id(self):  # pylint: disable=invalid-name
        """The id of the OU."""
        return self._data.get('Id')

    @property
    def name(self):
        """The name of the OU."""
        return self._data.get('Name')

    @property
    def arn(self):
        """The arn of the OU."""
        return self._data.get('Arn')


class ControlTowerOU:
    """Model the data of a Control Tower managed OU."""

    def __init__(self, control_tower, data):
        self.control_tower = control_tower
        self._data = data

    @property
    def create_date(self):
        """The date the ou was created in timestamp."""
        return self._data.get('CreateDate')

    @property
    def id(self):  # pylint: disable=invalid-name
        """OU ID."""
        return self._data.get('OrganizationalUnitId')

    @property
    def name(self):
        """The name of the OU."""
        return self._data.get('OrganizationalUnitName')

    @property
    def status(self):
        """The status of the OU."""
        return self._data.get('OrganizationalUnitStatus')

    @property
    def type(self):
        """The type of the OU."""
        return self._data.get('OrganizationalUnitType')

    @property
    def parent_ou_id(self):
        """The id of the parent OU."""
        return self._data.get('ParentOrganizationalUnitId')

    @property
    def parent_ou_name(self):
        """The name of the parent OU."""
        return self._data.get('ParentOrganizationalUnitName')

    def delete(self):
        """Deletes the ou.

        Returns:
            response (bool): True on success, False otherwise.

        """
        return self.control_tower.delete_organizational_unit(self.name)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File: entities.py
#
# Copyright 2020 Sayantan Khanra, Costas Tyfoxylos
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
#
"""
Main code for entities.

.. _Google Python Style Guide:
   http://google.github.io/styleguide/pyguide.html

"""

import logging
import json

from awsapilib.authentication import LoggerMixin

__author__ = '''Sayantan Khanra <skhanra@schubergphilis.com>, Costas Tyfoxylos <ctyfoxylos@schubergphilis.com>'''
__docformat__ = '''google'''
__date__ = '''18-05-2020'''
__copyright__ = '''Copyright 2020, Sayantan Khanra, Costas Tyfoxylos'''
__credits__ = ["Sayantan Khanra", "Costas Tyfoxylos"]
__license__ = '''MIT'''
__maintainer__ = '''Sayantan Khanra, Costas Tyfoxylos'''
__email__ = '''<skhanra@schubergphilis.com>, <ctyfoxylos@schubergphilis.com>'''
__status__ = '''Development'''  # "Prototype", "Development", "Production".

# This is the main prefix used for logging
LOGGER_BASENAME = '''entities'''
LOGGER = logging.getLogger(LOGGER_BASENAME)
LOGGER.addHandler(logging.NullHandler())


class Entity(LoggerMixin):  # pylint: disable=too-few-public-methods
    """The core entity."""

    def __init__(self, sso_instance, data):
        self._sso = sso_instance
        self._data = self._parse_data(data)

    def _parse_data(self, data):
        if not isinstance(data, dict):
            self.logger.error(f'Invalid data received :{data}')
            data = {}
        return data


class Group(Entity):
    """Models the group object of AWS SSO."""

    def __init__(self, sso_instance, data):
        super().__init__(sso_instance, data)
        self.url = f'{sso_instance.api_url}/userpool'

    @property
    def id(self):  # pylint: disable=invalid-name
        """The id of the group.

        Returns:
            id (str) : The id of the group

        """
        return self._data.get('GroupId')

    @property
    def name(self):
        """The name of the group.

        Returns:
            name (str) : The name of the group

        """
        return self._data.get('GroupName', '')

    @property
    def description(self):
        """The description of the group.

        Returns:
            description (str) : The description of the group

        """
        return self._data.get('Description', '')

    @property
    def users(self):
        """The users in the group.

        Returns:
            users (list): The users part of the group

        """
        content_payload = {'GroupId': self.id,
                           'MaxResults': 100}
        target = 'com.amazonaws.swbup.service.SWBUPService.ListMembersInGroup'
        for user in self._sso._get_paginated_results(content_payload=content_payload,  # pylint: disable=protected-access
                                                     path='userpool',
                                                     target='ListMembersInGroup',
                                                     amz_target=target,
                                                     object_group='Members',
                                                     url=self.url):
            yield self._sso.get_user_by_id(user.get('Id'))


class Account(Entity):
    """Models the Account object of AWS SSO."""

    @property
    def url(self):
        """Url for the account.

        Returns:
            url (str): The url of the account

        """
        return self._sso.endpoint_url

    @property
    def name(self):
        """The name of the application.

        Returns:
            name (str): The name of the application

        """
        return self._data.get('Name')

    @property
    def email(self):
        """The name of the application.

        Returns:
            email (str) : The name of the application

        """
        return self._data.get('Email')

    @property
    def id(self):  # pylint: disable=invalid-name
        """The id of the application.

        Returns:
            id (str): The id of the application

        """
        return self._data.get('Id')

    @property
    def arn(self):
        """The arn of the application.

        Returns:
            arn (str): The arn of the application

        """
        return self._data.get('Arn')

    @property
    def status(self):
        """The status of the application.

        Returns:
            status (str): The status of the application

        """
        return self._data.get('Status')

    def provision_saml_provider(self):
        """Creates the SAMl provider.

        Returns:
            arn (str): The arn of the SAMl provider

        """
        target = 'com.amazon.switchboard.service.SWBService.ProvisionSAMLProvider'
        payload = self._sso.get_api_payload(content_string={'applicationInstanceId': self.instance_id
                                                            },
                                            target='ProvisionSAMLProvider',
                                            path='/control/',
                                            x_amz_target=target)
        self.logger.debug('Trying to create saml provider for aws account with payload: %s', payload)
        response = self._sso.session.post(self.url, json=payload)
        if not response.ok:
            self.logger.error(response.text)
            return {}
        return response.json()

    @property
    def instance_id(self):
        """The instance id of the Account.

        Returns:
            instance_id (str): The instance id of the account

        """
        instance_id = self._retrieve_instance_id()
        if not instance_id:
            instance_id = self._provision_application_instance_for_aws_account()
        return instance_id

    def _provision_application_instance_for_aws_account(self):
        target = 'com.amazon.switchboard.service.SWBService.ProvisionApplicationInstanceForAWSAccount'
        payload = self._sso.get_api_payload(content_string={'accountId': self.id,
                                                            'accountEmail': self.email,
                                                            'accountName': self.name
                                                            },
                                            target='ProvisionApplicationInstanceForAWSAccount',
                                            path='/control/',
                                            x_amz_target=target)
        self.logger.debug('Trying to get instance id for aws account with payload: %s', payload)
        response = self._sso.session.post(self.url, json=payload)
        if not response.ok:
            self.logger.error(response.text)
            return None
        return response.json().get('applicationInstance', {}).get('instanceId', None)

    def _retrieve_instance_id(self):
        account_id = self.id
        target = 'com.amazon.switchboard.service.SWBService.GetApplicationInstanceForAWSAccount'
        payload = self._sso.get_api_payload(content_string={'awsAccountId': account_id},
                                            target='GetApplicationInstanceForAWSAccount',
                                            path='/control/',
                                            x_amz_target=target)
        self.logger.debug('Trying to get instance id for aws account with payload: %s', payload)
        response = self._sso.session.post(self.url, json=payload)
        if not response.ok:
            self.logger.error(response.text)
            return None
        return response.json().get('applicationInstance', {}).get('instanceId', None)

    @property
    def associated_profiles(self):
        """The associated profiles with the Account.

        Returns:
            associated_profiles (list): The profiles associated with the Account

        """
        target = 'com.amazon.switchboard.service.SWBService.ListAWSAccountProfiles'
        payload = self._sso.get_api_payload(content_string={'instanceId': self.instance_id},
                                            target='ListAWSAccountProfiles',
                                            path='/control/',
                                            x_amz_target=target)
        self.logger.debug('Trying to provision application profile for aws account with payload: %s', payload)
        response = self._sso.session.post(self.url, json=payload)
        if not response.ok:
            self.logger.error(response.text)
            return []
        return response.json().get('profileList', [])


class User(Entity):
    """Models the user object of SSO."""

    @property
    def url(self):
        """Url for the user.

        Returns:
            url (str): The url for the user

        """
        return f'{self._sso.api_url}/userpool'

    @property
    def status(self):
        """The status of the user.

        Returns:
            status (str): The status of the user

        """
        return self._data.get('Active')

    @property
    def created_at(self):
        """The date and time of the users's activation.

        Returns:
            created_at (datetime): The datetime object of when the user was activated

        """
        return self._data.get('Meta', {}).get('CreatedAt')

    @property
    def updated_at(self):
        """The date and time of the users's status change.

        Returns:
            updated_at (datetime): The datetime object of when the user had last changed status

        """
        return self._data.get('Meta', {}).get('UpdatedAt')

    @property
    def emails(self):
        """The date and time of the users's last password change.

        Returns:
            emails (datetime): The datetime object of when the user last changed password

        """
        return self._data.get('UserAttributes').get('emails', {}).get('ComplexListValue', '')

    @property
    def _name(self):
        return self._data.get('UserAttributes').get('name', {}).get('ComplexValue', {})

    @property
    def first_name(self):
        """The first name of the user.

        Returns:
            first_name (str): The first name of the user

        """
        return self._name.get('givenName', {}).get('StringValue', '')

    @property
    def last_name(self):
        """The last name of the user.

        Returns:
            last_name (str): The last name of the user

        """
        return self._name.get('familyName', {}).get('StringValue', '')

    @property
    def id(self):  # pylint: disable=invalid-name
        """The manager of the user.

        Returns:
            id (str): The manager of the user

        """
        return self._data.get('UserId')

    @property
    def name(self):
        """The manager of the user.

        Returns:
            name (str): The manager of the user

        """
        return self._data.get('UserName')

    @property
    def display_name(self):
        """The display name of the user.

        Returns:
            display_name (str): The display name of the user

        """
        return self._data.get('UserAttributes', {}).get('displayName', {}).get('StringValue')

    @property
    def groups(self):
        """The groups associated with the user.

        Returns:
            groups (list): The groups associated with the user

        """
        content_payload = {'UserId': self.id,
                           'MaxResults': 100}
        target = 'com.amazonaws.swbup.service.SWBUPService.ListGroupsForUser'
        for group in self._sso._get_paginated_results(content_payload=content_payload,  # pylint: disable=protected-access
                                                      path='userpool',
                                                      target='ListGroupsForUser',
                                                      amz_target=target,
                                                      object_group='Groups',
                                                      url=self.url):
            yield self._sso.get_group_by_id(group.get('GroupId'))


class PermissionSet(Entity):
    """Models the permission set object of SSO."""

    @property
    def url(self):
        """Url of the permission set.

        Returns:
            url (str): The url of the permission set

        """
        return self._sso.endpoint_url

    @property
    def description(self):
        """The description of the permission set.

        Returns:
            description (str): The description of the permission set

        """
        return self._data.get('Description')

    @property
    def id(self):  # pylint: disable=invalid-name
        """The id of the permission set.

        Returns:
            id (str): The id of the permission set

        """
        return self._data.get('Id')

    @property
    def name(self):
        """The name of the permission set.

        Returns:
            name (str): The name of the permission set

        """
        return self._data.get('Name')

    @property
    def ttl(self):
        """The ttl of the permission set.

        Returns:
            ttl (str): The ttl of the permission set

        """
        return self._data.get('ttl')

    @property
    def creation_date(self):
        """The creation date of the permission set.

        Returns:
            creation_date (str): The creation date of the permission set

        """
        return self._data.get('CreationDate')

    @property
    def relay_state(self):
        """The relay_state of the permission_set.

        Returns:
            relay_state (str): The relayState of the permission_set

        """
        return self._data.get('relayState')

    @property
    def permission_policy(self):
        """The permission policy of the permission_set.

        Returns:
            permission_policy (dict): The permission policy of the permission_set

        """
        target = 'com.amazon.switchboard.service.SWBService.GetPermissionsPolicy'
        content_string = {'permissionSetId': self.id}
        payload = self._sso.get_api_payload(content_string=content_string,
                                            target='GetPermissionsPolicy',
                                            path='/control/',
                                            x_amz_target=target)
        self.logger.debug('Getting permission policy for permission_set with payload of %s:', payload)
        response = self._sso.session.post(self.url, json=payload)
        if not response.ok:
            self.logger.error(response.text)
            return None
        return response.json()

    @property
    def provisioned_accounts(self):
        """The provisioned accounts with the permission set.

        Returns:
            list: Accounts provisioned with the permission set

        """
        content_payload = {'permissionSetId': self.id,
                           'onlyOutOfSync': 'false'}
        target = 'com.amazon.switchboard.service.SWBService.ListAccountsWithProvisionedPermissionSet'
        for account_id in self._sso._get_paginated_results(content_payload=content_payload,  # pylint: disable=protected-access
                                                           path='control',
                                                           target='ListAccountsWithProvisionedPermissionSet',
                                                           amz_target=target,
                                                           object_group='accountIds',
                                                           next_token_marker='marker',
                                                           url=self._sso.endpoint_url):
            yield self._sso.get_account_by_id(account_id)

    def assign_custom_policy_to_permission_set(self, policy_document):
        """Assign Custom policy to a permission_set.

        Args:
            permission_set_name: The name of the permission_set .
            policy_document: The policy for the permission_set
        Returns:
            Bool:  True or False

        """
        content_string = {'permissionSetId': self.id,
                          'policyDocument': json.dumps(policy_document)}
        target = 'com.amazon.switchboard.service.SWBService.PutPermissionsPolicy'
        payload = self._sso.get_api_payload(content_string=content_string,
                                            target='PutPermissionsPolicy',
                                            path='/control/',
                                            x_amz_target=target)
        self.logger.debug('Assigning custom policy to permission set with payload %s:', payload)
        response = self._sso.session.post(self.url, json=payload)
        if not response.ok:
            self.logger.error(response.text)
        else:
            if self.provisioned_accounts:  # pylint: disable=using-constant-test
                for account in self.provisioned_accounts:
                    self.logger.debug('Updating associated account %s', account.name)
                    self._sso._provision_application_profile_for_aws_account_instance(self.name, account.name)  # pylint: disable=protected-access
        return response.ok

    def delete_custom_policy_from_permission_set(self):
        """Assign Custom policy to a permission_set.

        Returns:
            Bool:  True or False

        """
        content_string = {'permissionSetId': self.id}
        target = 'com.amazon.switchboard.service.SWBService.DeletePermissionsPolicy'
        payload = self._sso.get_api_payload(content_string=content_string,
                                            target='DeletePermissionsPolicy',
                                            path='/control/',
                                            x_amz_target=target)
        response = self._sso.session.post(self.url, json=payload)
        if not response.ok:
            self.logger.error(response.text)
        else:
            if self.provisioned_accounts:  # pylint: disable=using-constant-test
                for account in self.provisioned_accounts:
                    self.logger.debug('Updating associated account %s', account.name)
                    self._sso._provision_application_profile_for_aws_account_instance(self.name, account.name)  # pylint: disable=protected-access
        return response.ok

    def update(self, description=' ', relay_state='', ttl=''):
        """The relayState of the permission_set.

        Args:
            description: Description for the permission set
            relay_state: The relay state for the permission set.
                                 https://docs.aws.amazon.com/singlesignon/latest/userguide/howtopermrelaystate.html
            ttl: session duration

        Returns:
            bool: True or False

        """
        content_string = {'permissionSetId': self.id,
                          'description': description if description != ' ' else self.description,
                          'ttl': ttl if ttl else self.ttl,
                          'relayState': relay_state if relay_state else self.relay_state}
        target = 'com.amazon.switchboard.service.SWBService.UpdatePermissionSet'
        payload = self._sso.get_api_payload(content_string=content_string,
                                            target='UpdatePermissionSet',
                                            path='/control/',
                                            x_amz_target=target)
        self.logger.debug('Posting to url %s payload of %s:', self.url, payload)
        response = self._sso.session.post(self.url, json=payload)
        if not response.ok:
            self.logger.error(response.text)
        return response.ok



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File: authentication.py
#
# Copyright 2020 Costas Tyfoxylos
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
#

"""
Main code for authentication.

.. _Google Python Style Guide:
   http://google.github.io/styleguide/pyguide.html

"""
import json
import logging
import urllib

from copy import deepcopy
from dataclasses import dataclass

import boto3
import botocore
import requests

from bs4 import BeautifulSoup as Bfs

from .authenticationexceptions import NoSigninTokenReceived, InvalidCredentials, ExpiredCredentials

__author__ = '''Costas Tyfoxylos <ctyfoxylos@schubergphilis.com>'''
__docformat__ = '''google'''
__date__ = '''11-03-2020'''
__copyright__ = '''Copyright 2020, Costas Tyfoxylos'''
__credits__ = ["Costas Tyfoxylos"]
__license__ = '''MIT'''
__maintainer__ = '''Costas Tyfoxylos'''
__email__ = '''<ctyfoxylos@schubergphilis.com>'''
__status__ = '''Development'''  # "Prototype", "Development", "Production".

# This is the main prefix used for logging
LOGGER_BASENAME = '''authentication'''
LOGGER = logging.getLogger(LOGGER_BASENAME)
LOGGER.addHandler(logging.NullHandler())

DEFAULT_REGION = 'eu-west-1'


@dataclass
class FilterCookie:
    """Object modeling a cookie for filtering."""

    name: str
    domain: str = ''
    exact_match: bool = False


@dataclass
class CsrfTokenData:
    """Object modeling the data required for csrf token filtering."""

    entity_type: str
    attributes: dict
    attribute_value: str
    headers_name: str


@dataclass
class Domains:
    """Dataclass holding the domains required for authenticating."""

    region: str
    root: str = 'aws.amazon.com'
    sign_in: str = f'signin.{root}'
    console: str = f'console.{root}'

    @property
    def regional_console(self):
        """The domain of the regional console.

        Returns:
            regional_console (str): The regional console domain.

        """
        return f'{self.region}.console.{self.root}'


@dataclass
class Urls:
    """Dataclass holding the urls required for authenticating."""

    region: str
    scheme: str = 'https://'
    root_domain: str = 'aws.amazon.com'
    root: str = f'{scheme}{root_domain}'
    sign_in: str = f'{scheme}signin.{root_domain}'
    console: str = f'{scheme}console.{root_domain}'
    federation: str = f'{sign_in}/federation'

    @property
    def regional_console(self):
        """The url of the regional console.

        Returns:
            regional_console (str): The regional console url.

        """
        return f'{self.scheme}{self.region}.console.{self.root_domain}'

    @property
    def regional_single_sign_on(self):
        """The url of the regional single sign on.

        Returns:
            regional_single_sign_on (str): The regional single sign on url.

        """
        return f'{self.scheme}{self.region}.console.{self.root_domain}/singlesignon'

    @property
    def regional_control_tower(self):
        """The url of the regional control tower service.

        Returns:
            regional_control_tower (str): The regional control tower on url.

        """
        return f'{self.scheme}{self.region}.console.{self.root_domain}/controltower'

    @property
    def regional_relay_state(self):
        """The regional relay state url.

        Returns:
            relay_state (str): The regional relay state url.

        """
        return f'{self.regional_console}home?region={self.region}#'


class LoggerMixin:  # pylint: disable=too-few-public-methods
    """Logger."""

    @property
    def logger(self):
        """Exposes the logger to be used by objects using the Mixin.

        Returns:
            logger (logger): The properly named logger.

        """
        return logging.getLogger(f'{LOGGER_BASENAME}.{self.__class__.__name__}')


class Authenticator(LoggerMixin):   # pylint: disable=too-many-instance-attributes
    """Interfaces with aws authentication mechanisms, providing pre signed urls, or authenticated sessions."""

    def __init__(self, arn, session_duration=3600, region=None):
        self.arn = arn
        self.session_duration = session_duration
        self._session = requests.Session()
        self._sts_connection = boto3.client('sts', region_name=region)
        self.region = region or self._get_region()
        self._assumed_role = self._get_assumed_role(arn)
        self.urls = Urls(self.region)
        self.domains = Domains(self.region)

    def _get_region(self):
        region = self._sts_connection._client_config.region_name  # pylint: disable=protected-access
        return region if not region == 'aws-global' else DEFAULT_REGION

    def _get_assumed_role(self, arn):
        self.logger.debug('Trying to assume role "%s".', arn)
        try:
            return self._sts_connection.assume_role(RoleArn=arn,
                                                    RoleSessionName="AssumeRoleSession",
                                                    DurationSeconds=self.session_duration)
        except botocore.exceptions.ParamValidationError as error:
            raise ValueError('The arn you provided is incorrect: {}'.format(error)) from None
        except (botocore.exceptions.NoCredentialsError, botocore.exceptions.ClientError) as error:
            raise InvalidCredentials(error) from None

    @property
    def session_credentials(self):
        """Valid credentials for a session.

        Returns:
            credentials (dict): A properly structured dictionary of session credentials.

        """
        payload = {'sessionId': 'AccessKeyId',
                   'sessionKey': 'SecretAccessKey',
                   'sessionToken': 'SessionToken'}
        return self._get_credentials(payload)

    @property
    def assumed_role_credentials(self):
        """Valid credentials for an assumed session.

        Returns:
            credentials (dict): A properly structured dictionary of an assumed session credentials.

        """
        payload = {'aws_access_key_id': 'AccessKeyId',
                   'aws_secret_access_key': 'SecretAccessKey',
                   'aws_session_token': 'SessionToken'}
        return self._get_credentials(payload)

    def _get_credentials(self, payload):
        self.logger.debug('Getting credentials from assumed role object.')
        credentials = self._assumed_role.get('Credentials')
        self.logger.debug('Building payload.')
        payload_ = {key: credentials.get(value)
                    for key, value in payload.items()}
        return payload_

    def _get_signin_token(self):  # we can pass a duration here.
        self.logger.debug('Trying to get signin token.')
        params = {'Action': 'getSigninToken',
                  # 'SessionDuration': str(duration),
                  'Session': json.dumps(self.session_credentials)}
        response = requests.get(self.urls.federation, params=params)
        if all([response.status_code == 401, response.text == 'Token Expired']):
            try:
                self._assumed_role = self._get_assumed_role(self.arn)
                return self._get_signin_token()
            except InvalidCredentials:
                self.logger.error('The credentials on the environment do not provide access for session refresh.')
                raise
        if response.ok:
            return response.json().get('SigninToken')
        raise NoSigninTokenReceived(response.status_code, response.text)

    def get_signed_url(self, domain='Example.com', destination=None):
        """Returns a pre signed url that is authenticated.

        Args:
            domain (str): The domain to request the session as.
            destination (str): The service to redirect to after successful redirection.


        Returns:
            url (str): An authenticated pre signed url.

        """
        params = {'Action': 'login',
                  'Issuer': domain,
                  'Destination': destination or self.urls.console,
                  'SigninToken': self._get_signin_token()}
        return f'{self.urls.federation}?{urllib.parse.urlencode(params)}'

    @staticmethod
    def _filter_cookies(cookies, filters=None):
        result_cookies = []
        for filter_ in filters:
            for cookie in cookies:
                conditions = [cookie.name == filter_.name]
                if filter_.exact_match:
                    conditions.extend([filter_.domain == f'{cookie.domain}{cookie.path}'])
                elif filter_.domain:
                    conditions.extend([filter_.domain in f'{cookie.domain}{cookie.path}'])
                if all(conditions):
                    result_cookies.append(cookie)
        return result_cookies

    @staticmethod
    def _cookies_to_dict(cookies):
        return {cookie.name: cookie.value for cookie in cookies}

    @staticmethod
    def _header_cookie_from_cookies(cookies):
        return '; '.join([f'{key}={value}'
                          for key, value in Authenticator._cookies_to_dict(cookies).items()])

    @property
    def _default_headers(self):
        return deepcopy({'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,*/*;q=0.8',
                         'Accept-Encoding': 'gzip, deflate, br',
                         'Accept-Language': 'en-US,en;q=0.5',
                         'Connection': 'keep-alive',
                         'Upgrade-Insecure-Requests': '1',
                         'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10.14; rv:73.0) '
                                       'Gecko/20100101 Firefox/73.0'})

    @property
    def _standard_cookies(self):
        return [FilterCookie('aws-account-data'),
                FilterCookie('aws-ubid-main'),
                FilterCookie('aws-userInfo'),
                FilterCookie('awsc-actm'),
                FilterCookie('awsm-vid')]

    def _get_response(self, url, params=None, extra_cookies=None, headers=None):
        extra_cookies = extra_cookies or []
        headers = headers or {}
        cookies_to_filter = self._standard_cookies + extra_cookies
        headers.update(self._default_headers)
        cookies = self._filter_cookies(self._session.cookies, cookies_to_filter)
        headers['Cookie'] = self._header_cookie_from_cookies(cookies)
        arguments = {'url': url,
                     'headers': headers,
                     'cookies': self._cookies_to_dict(cookies),
                     'allow_redirects': False}
        if params:
            arguments.update({'params': params})
        self.logger.debug('Getting url :%s with arguments : %s', url, arguments)
        response = requests.get(**arguments)
        if not response.ok:
            try:
                error_response = Bfs(response.text, features='html.parser')
                error_title = error_response.title.string.strip()
                err_msg = error_response.find('div', {'id': 'content'}).find('p').string
            except AttributeError:
                raise ValueError('Response received: %s' % response.text)
            if all([response.status_code == 400, error_title == 'Credentials expired']):
                raise ExpiredCredentials(response.status_code, err_msg)
            raise ValueError('Response received: %s' % response.text)
        self._debug_response(response, cookies)
        self._session.cookies.update(response.cookies)
        return response

    @staticmethod
    def _query_to_params(query):
        query_lines = query.split('&') if query else []
        return {line.split('=')[0]: line.split('=')[1] for line in query_lines}

    def _debug_response(self, response, cookies):
        params = self._query_to_params(urllib.parse.urlparse(response.request.url)[4])
        self.logger.debug('URL : %s', response.request.url)
        if params:
            self.logger.debug('Params : ')
            for key, value in params.items():
                self.logger.debug('\t%s : %s', key, value)
        self.logger.debug('Response status : %s', response.status_code)
        self.logger.debug('\tRequest Headers :')
        for name, value in dict(sorted(response.request.headers.items(), key=lambda x: x[0].lower())).items():
            self.logger.debug('\t\t%s : %s', name, value)
        self.logger.debug('\tRequest Cookies :')
        for cookie in sorted(cookies, key=lambda x: x.name.lower()):
            self.logger.debug('\t\t%s (domain:%s) : %s', cookie.name, cookie.domain, cookie.value)
        self.logger.debug('\tResponse Headers :')
        for name, value in dict(sorted(response.headers.items(), key=lambda x: x[0].lower())).items():
            self.logger.debug('\t\t%s : %s', name, value)
        self.logger.debug('\tResponse Cookies :')
        for name, value in dict(sorted(response.cookies.items(), key=lambda x: x[0].lower())).items():
            self.logger.debug('\t\t%s : %s', name, value)
        self.logger.debug('Session Cookies :')
        for cookie in sorted(self._session.cookies, key=lambda x: x.name.lower()):
            self.logger.debug('\t%s (domain:%s%s) : %s', cookie.name, cookie.domain, cookie.path, cookie.value)

    def get_control_tower_authenticated_session(self):
        """Authenticates to control tower and returns an authenticated session.

        Returns:
            session (requests.Session): An authenticated session with headers and cookies set.

        """
        service = 'controltower'
        self._session.get(self.get_signed_url())
        url = f'{self.urls.regional_console}/{service}/home'
        host = urllib.parse.urlparse(url)[1]
        self.logger.debug('Setting host to: %s', host)
        self._get_response(url,
                           params={'region': self.region},
                           extra_cookies=[FilterCookie('JSESSIONID'),
                                          FilterCookie('aws-userInfo-signed')])
        hash_args = self._get_response(url,
                                       params={'state': 'hashArgs#'},
                                       extra_cookies=[FilterCookie('JSESSIONID', self.urls.regional_console),
                                                      FilterCookie('aws-userInfo-signed',),
                                                      FilterCookie('aws-creds-code-verifier', self.urls.regional_console
                                                                   )])
        oauth = self._get_response(hash_args.headers.get('Location'),
                                   extra_cookies=[FilterCookie('JSESSIONID', self.urls.regional_console),
                                                  FilterCookie('aws-creds', self.domains.sign_in),
                                                  FilterCookie('aws-userInfo-signed', ),
                                                  FilterCookie('aws-creds-code-verifier', f'/{service}')],)
        oauth_challenge = self._get_response(oauth.headers.get('Location'),
                                             extra_cookies=[FilterCookie('JSESSIONID', self.urls.regional_console),
                                                            FilterCookie('aws-userInfo-signed',),
                                                            FilterCookie('aws-creds', self.domains.sign_in),
                                                            FilterCookie('aws-creds-code-verifier', f'/{service}')])
        self._get_response(oauth_challenge.headers.get('Location'),
                           extra_cookies=[FilterCookie('aws-creds', f'/{service}'),
                                          FilterCookie('JSESSIONID', host),
                                          FilterCookie('aws-userInfo-signed')])
        dashboard = self._get_response(url,
                                       params={'region': self.region},
                                       extra_cookies=[FilterCookie('aws-creds', f'/{service}'),
                                                      FilterCookie('JSESSIONID', host),
                                                      FilterCookie('aws-consoleInfo'),
                                                      FilterCookie('aws-userInfo-signed')])
        csrf_token_data = CsrfTokenData('meta', {'name': 'awsc-csrf-token'}, 'content', 'X-CSRF-TOKEN')
        extra_cookies = [FilterCookie('JSESSIONID', self.domains.regional_console),
                         FilterCookie('aws-creds', f'{self.domains.regional_console}/{service}')]
        return self._get_session_from_console(dashboard, csrf_token_data, extra_cookies)

    def get_sso_authenticated_session(self):
        """Authenticates to Single Sign On and returns an authenticated session.

        Returns:
            session (requests.Session): An authenticated session with headers and cookies set.

        """
        service = 'singlesignon'
        url = f'{self.urls.regional_console}/{service}/home?region={self.region}#/dashboard'
        self._get_response(self.get_signed_url())
        host = urllib.parse.urlparse(url)[1]
        self.logger.debug('Setting host to: %s', host)
        self._get_response(url, extra_cookies=[FilterCookie('JSESSIONID', f'/{service}')])
        hash_args = self._get_response(url,
                                       params={'state': 'hashArgs#'},
                                       extra_cookies=[FilterCookie('JSESSIONID', f'/{service}'),
                                                      FilterCookie('aws-userInfo-signed', )])
        oauth = self._get_response(hash_args.headers.get('Location'),
                                   extra_cookies=[FilterCookie('aws-creds', self.domains.sign_in),
                                                  FilterCookie('aws-userInfo-signed', )])
        oauth_challenge = self._get_response(oauth.headers.get('Location'),
                                             extra_cookies=[FilterCookie('JSESSIONID', self.urls.regional_console),
                                                            FilterCookie('aws-userInfo-signed', ),
                                                            FilterCookie('aws-creds-code-verifier', f'/{service}')
                                                            ])
        dashboard = self._get_response(oauth_challenge.headers.get('Location'),
                                       extra_cookies=[FilterCookie('aws-creds', f'/{service}'),
                                                      FilterCookie('JSESSIONID', host)])
        csrf_token_data = CsrfTokenData('meta', {'name': 'awsc-csrf-token'}, 'content', 'X-CSRF-TOKEN')
        extra_cookies = [FilterCookie('JSESSIONID', self.domains.regional_console),
                         FilterCookie('aws-creds', f'{self.domains.regional_console}/{service}')]
        return self._get_session_from_console(dashboard, csrf_token_data, extra_cookies)

    def get_billing_authenticated_session(self):
        """Authenticates to billing and returns an authenticated session.

        Returns:
            session (requests.Session): An authenticated session with headers and cookies set.

        """
        service = 'billing'
        url = f'{self.urls.console}/{service}/home?region={self.region}'
        self._get_response(self.get_signed_url())
        host = urllib.parse.urlparse(url)[1]
        self.logger.debug('Setting host to: %s', host)
        self._get_response(url, extra_cookies=[FilterCookie('JSESSIONID', f'/{service}'),
                                               FilterCookie('aws-userInfo-signed', )])
        hash_args = self._get_response(url,
                                       params={'state': 'hashArgs#'},
                                       extra_cookies=[FilterCookie('JSESSIONID', f'/{service}'),
                                                      FilterCookie('aws-userInfo-signed', ),
                                                      FilterCookie('aws-creds-code-verifier', f'/{service}')])
        oauth = self._get_response(hash_args.headers.get('Location'),
                                   extra_cookies=[FilterCookie('aws-creds', self.domains.sign_in),
                                                  FilterCookie('aws-userInfo-signed', )])
        oauth_challenge = self._get_response(oauth.headers.get('Location'),
                                             extra_cookies=[FilterCookie('JSESSIONID', self.urls.regional_console),
                                                            FilterCookie('aws-userInfo-signed', ),
                                                            FilterCookie('aws-creds-code-verifier', f'/{service}')
                                                            ])
        dashboard = self._get_response(oauth_challenge.headers.get('Location'),
                                       extra_cookies=[FilterCookie('aws-creds', f'/{service}'),
                                                      FilterCookie('JSESSIONID', host)])
        csrf_token_data = CsrfTokenData('input', {'id': 'xsrfToken'}, 'value', 'x-awsbc-xsrf-token')
        extra_cookies = [FilterCookie('aws-creds', '/billing')]
        return self._get_session_from_console(dashboard, csrf_token_data, extra_cookies)

    def _get_session_from_console(self, console_page_response, csrf_token_data, extra_cookies=None):
        soup = Bfs(console_page_response.text, features='html.parser')
        try:
            csrf_token = soup.find(csrf_token_data.entity_type,
                                   csrf_token_data.attributes).attrs.get(csrf_token_data.attribute_value)
        except AttributeError:
            raise ValueError('Response received: %s' % console_page_response.text)
        if not csrf_token:
            raise NoSigninTokenReceived('Unable to retrieve csrf token.')
        session = requests.Session()
        cookies_to_filter = self._standard_cookies + extra_cookies if extra_cookies else []
        cookies = self._filter_cookies(self._session.cookies, cookies_to_filter)
        session.headers.update(self._default_headers)
        session.headers.update({'Cookie': self._header_cookie_from_cookies(cookies),
                                csrf_token_data.headers_name: csrf_token})
        for cookie in cookies:
            session.cookies.set_cookie(cookie)
        return session



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File: sso.py
#
# Copyright 2020 Sayantan Khanra, Costas Tyfoxylos
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
#

"""
Main code for sso.

.. _Google Python Style Guide:
   http://google.github.io/styleguide/pyguide.html

"""

import copy
import logging
import json
from awsapilib.authentication import (Authenticator,
                                      LoggerMixin,
                                      Urls)
from .ssoexceptions import (UnsupportedTarget,
                            NoPermissionSet,
                            NoAccount,
                            NoGroup,
                            NoProfileID,
                            NoUser)
from .entities import (Group,
                       User,
                       Account,
                       PermissionSet)

__author__ = '''Sayantan Khanra <skhanra@schubergphilis.com>, Costas Tyfoxylos <ctyfoxylos@schubergphilis.com>'''
__docformat__ = '''google'''
__date__ = '''18-05-2020'''
__copyright__ = '''Copyright 2020, Sayantan Khanra, Costas Tyfoxylos'''
__credits__ = ["Sayantan Khanra", "Costas Tyfoxylos"]
__license__ = '''MIT'''
__maintainer__ = '''Sayantan Khanra, Costas Tyfoxylos'''
__email__ = '''<skhanra@schubergphilis.com>, <ctyfoxylos@schubergphilis.com>'''
__status__ = '''Development'''  # "Prototype", "Development", "Production".

# This is the main prefix used for logging
LOGGER_BASENAME = '''sso'''
LOGGER = logging.getLogger(LOGGER_BASENAME)
LOGGER.addHandler(logging.NullHandler())

SUPPORTED_TARGETS = ['GetUserPoolInfo',
                     'SearchGroups',
                     'ProvisionApplicationInstanceForAWSAccount',
                     'ListPermissionSets',
                     'GetApplicationInstanceForAWSAccount',
                     'ProvisionApplicationProfileForAWSAccountInstance',
                     'AssociateProfile',
                     'ListAWSAccountProfiles',
                     'DisassociateProfile',
                     'SearchUsers',
                     'ListMembersInGroup',
                     'ListGroupsForUser',
                     'CreatePermissionSet',
                     'PutPermissionsPolicy',
                     'GetPermissionsPolicy',
                     'ListAccountsWithProvisionedPermissionSet',
                     'UpdatePermissionSet',
                     'listAccounts',
                     'DeletePermissionSet',
                     'DeletePermissionsPolicy',
                     'ProvisionApplicationInstanceForAWSAccount',
                     'ProvisionSAMLProvider']


class Sso(LoggerMixin):  # pylint: disable=too-many-public-methods
    """Models AWS SSO."""

    API_CONTENT_TYPE = 'application/json; charset=UTF-8'
    API_CONTENT_ENCODING = 'amz-1.0'
    DEFAULT_AWS_REGION = 'eu-west-1'

    def __init__(self, arn, region=None):
        self.aws_authenticator = Authenticator(arn, region=region)
        self._urls = Urls(self.aws_region)
        self.session = self._get_authenticated_session()
        self._directory_id = None

    @property
    def relay_state(self):
        """The relay state of the SSO.

        Returns:
            relay_state (str): The relay state of sso.

        """
        return self._urls.regional_relay_state

    @property
    def api_url(self):
        """The url of the api for sso.

        Returns:
            api_url (str): The url of the api for sso.

        """
        return f'{self._urls.regional_single_sign_on}/api'

    @property
    def endpoint_url(self):
        """The url of the api endpoint for sso.

        Returns:
            endpoint_url (str): The url of the api endpoint for sso.

        """
        return f'{self.api_url}/peregrine'

    @property
    def aws_region(self):
        """Aws Console Region.

        Returns:
            region (str): The region of the console.

        """
        return self.aws_authenticator.region

    def get_api_payload(self,  # pylint: disable=too-many-arguments
                        content_string,
                        target,
                        method='POST',
                        params=None,
                        path='/',
                        content_type=None,
                        content_encoding=None,
                        x_amz_target='',
                        region=None):
        """Generates the payload for calling the AWS SSO APIs.

        Returns:
            payload (dict): Returns a deepcopy object of the payload

        """
        target = self._validate_target(target)
        payload = {'contentString': json.dumps(content_string),
                   'headers': {'Content-Type': content_type or self.API_CONTENT_TYPE,
                               'Content-Encoding': content_encoding or self.API_CONTENT_ENCODING,
                               'X-Amz-Target': x_amz_target},
                   'method': method,
                   'operation': target,
                   'params': params or {},
                   'path': path,
                   'region': region or self.DEFAULT_AWS_REGION}
        return copy.deepcopy(payload)

    @staticmethod
    def _validate_target(target):
        if target not in SUPPORTED_TARGETS:
            raise UnsupportedTarget(target)
        return target

    def _get_authenticated_session(self):
        return self.aws_authenticator.get_sso_authenticated_session()

    @property
    def directory_id(self):
        """The external/internal directory id configured with aws sso.

        Returns:
           str: The id of directory configured in SSO

        """
        if self._directory_id is None:
            payload = self.get_api_payload(content_string={},
                                           target='GetUserPoolInfo',
                                           path='/userpool/',
                                           x_amz_target='com.amazonaws.swbup.service.SWBUPService.GetUserPoolInfo',
                                           region=self.aws_region)
            self.logger.debug('Trying to get directory id for sso with payload: %s', payload)
            response = self.session.post(f'{self.api_url}/userpool', json=payload)
            if not response.ok:
                raise ValueError(response.text)
            self._directory_id = response.json().get('DirectoryId')
        return self._directory_id

    @property
    def accounts(self):
        """The aws accounts in sso.

        Returns:
            accounts (generator): The accounts configured in SSO

        """
        headers = {'Content-Type': 'application/x-amz-json-1.1',
                   'Content-Encoding': 'amz-1.0',
                   'X-Amz-Target': 'AWSOrganizationsV20161128.ListAccounts',
                   'X-Amz-User-Agent': 'aws-sdk-js/2.152.0 promise'}
        return self._get_paginated_results(content_payload={},
                                           path='',
                                           target='listAccounts',
                                           amz_target='AWSOrganizationsV20161128.ListAccounts',
                                           region='us-east-1',
                                           object_type=Account,
                                           object_group='Accounts',
                                           url=f'{self.api_url}/organizations',
                                           headers=headers)

    @property
    def users(self):
        """The users configured in SSO.

        Returns:
            users (generator): The users configured in SSO

        """
        content_payload = {'IdentityStoreId': self.directory_id,
                           'MaxResults': 50}
        return self._get_paginated_results(content_payload,
                                           path='identitystore',
                                           target='SearchUsers',
                                           amz_target='com.amazonaws.identitystore.AWSIdentityStoreService.SearchUsers',
                                           region=self.aws_region,
                                           object_type=User,
                                           object_group='Users')

    @property
    def groups(self):
        """The groups configured in SSO.

        Returns:
            groups (generator): The groups configured in SSO

        """
        content_payload = {'SearchString': '*',
                           'SearchAttributes': ['GroupName'],
                           'MaxResults': 100}
        return self._get_paginated_results(content_payload,
                                           path='userpool',
                                           target='SearchGroups',
                                           amz_target='com.amazonaws.swbup.service.SWBUPService.SearchGroups',
                                           region=self.aws_region,
                                           object_type=Group,
                                           object_group='Groups')

    @property
    def permission_sets(self):
        """The permission_sets configured in SSO.

        Returns:
            permission_sets (generator): The permission sets configured in SSO

        """
        return self._get_paginated_results(content_payload={},
                                           path='control',
                                           target='ListPermissionSets',
                                           amz_target='com.amazon.switchboard.service.SWBService.ListPermissionSets',
                                           region=self.aws_region,
                                           object_type=PermissionSet,
                                           object_group='permissionSets',
                                           url=self.endpoint_url,
                                           next_token_marker='marker')

    def get_user_by_name(self, user_name):
        """The user configured in SSO.

        Returns:
            user (User): The User object

        """
        return next((user for user in self.users if user.name == user_name), None)

    def get_user_by_id(self, user_id):
        """The user configured in SSO.

        Returns:
            user (User): The User object

        """
        return next((user for user in self.users if user.id == user_id), None)

    def get_group_by_name(self, group_name):
        """The group configured in SSO.

        Returns:
            group (Group): The Group object

        """
        return next((group for group in self.groups if group.name == group_name), None)

    def get_group_by_id(self, group_id):
        """The group configured in SSO.

        Returns:
            group (Group): The Group object

        """
        return next((group for group in self.groups if group.id == group_id), None)

    def get_account_by_name(self, account_name):
        """The account configured in SSO.

        Returns:
            account (Account): The Account object

        """
        return next((account for account in self.accounts if account.name == account_name), None)

    def get_account_by_id(self, account_id):
        """The account configured in SSO.

        Returns:
            account (Account): The Account object

        """
        return next((account for account in self.accounts if account.id == account_id), None)

    def get_permission_set_by_name(self, permission_set_name):
        """The permission-set configured in SSO.

        Returns:
            permission_set (PermissionSet): The PermissionSet object

        """
        return next((permission_set for permission_set in self.permission_sets
                     if permission_set.name == permission_set_name), None)

    def _provision_application_profile_for_aws_account_instance(self, permission_set_name, account_name):
        method = 'ProvisionApplicationProfileForAWSAccountInstance'
        permission_set = self.get_permission_set_by_name(permission_set_name)
        if not permission_set:
            raise NoPermissionSet(permission_set_name)
        account = self.get_account_by_name(account_name)
        if not account:
            raise NoAccount(account_name)
        payload = self.get_api_payload(content_string={'permissionSetId': permission_set.id,
                                                       'instanceId': account.instance_id},
                                       target=method,
                                       path='/control/',
                                       x_amz_target=f'com.amazon.switchboard.service.SWBService.{method}',
                                       region=self.aws_region)
        self.logger.debug('Trying to provision application profile for aws account with payload: %s', payload)
        response = self.session.post(self.endpoint_url, json=payload)
        if not response.ok:
            raise ValueError(response.text)
        return response.json().get('applicationProfile', {}).get('profileId', '')

    def _get_aws_account_profile_for_permission_set(self, account_name, permission_set_name):
        account = self.get_account_by_name(account_name)
        if not account:
            raise NoAccount(account_name)
        return next((profile for profile in account.associated_profiles
                     if profile.get('name') == permission_set_name), None)

    def associate_group_to_account(self, group_name, account_name, permission_set_name):
        """Associates a group with an account with proper permissions.

        Args:
            group_name: The name of the group to be assigned.
            account_name: Name of the account to which the group will be assigned
            permission_set_name: the Permission Set the group will have on the account

        Returns:
            bool: True or False

        """
        group = self.get_group_by_name(group_name)
        if not group:
            raise NoGroup(group_name)
        account = self.get_account_by_name(account_name)
        if not account:
            raise NoAccount(account_name)
        profile_id = self._provision_application_profile_for_aws_account_instance(permission_set_name, account_name)
        directory_id = self.directory_id
        content_string = {'accessorId': group.id,
                          'accessorType': 'GROUP',
                          'accessorDisplay': {'groupName': group_name},
                          'instanceId': account.instance_id,
                          'profileId': profile_id,
                          'directoryType': 'UserPool',
                          'directoryId': directory_id}
        payload = self.get_api_payload(content_string=content_string,
                                       target='AssociateProfile',
                                       path='/control/',
                                       x_amz_target='com.amazon.switchboard.service.SWBService.AssociateProfile',
                                       region=self.aws_region)
        self.logger.debug('Trying to assign groups to aws account with payload: %s', payload)
        response = self.session.post(self.endpoint_url, json=payload)
        if not response.ok:
            self.logger.error('Received :%s', response.text)
        return response.ok

    def disassociate_group_from_account(self, group_name, account_name, permission_set_name):
        """Disassociates a group with an account with proper permissions.

        Args:
            group_name: The name of the group to be assigned.
            account_name: Name of the account to which the group will be assigned
            permission_set_name: the Permission Set the group will have on the account

        Returns:
            bool: True or False

        """
        group = self.get_group_by_name(group_name)
        if not group:
            raise NoGroup(group_name)
        account = self.get_account_by_name(account_name)
        if not account:
            raise NoAccount(account_name)
        profile_id = self._get_aws_account_profile_for_permission_set(account_name,
                                                                      permission_set_name).get('profileId')
        if not profile_id:
            raise NoProfileID(f'{account_name}:{permission_set_name}')
        content_string = {'accessorId': group.id,
                          'accessorType': 'GROUP',
                          'accessorDisplay': {"groupName": group_name},
                          'instanceId': account.instance_id,
                          'profileId': profile_id,
                          'directoryType': 'UserPool',
                          'directoryId': self.directory_id}
        payload = self.get_api_payload(content_string=content_string,
                                       target='DisassociateProfile',
                                       path='/control/',
                                       x_amz_target='com.amazon.switchboard.service.SWBService.DisassociateProfile',
                                       region=self.aws_region)
        self.logger.debug('Trying to disassociate group from aws account with payload: %s', payload)
        response = self.session.post(self.endpoint_url, json=payload)
        if not response.ok:
            self.logger.error('Received :%s', response.text)
        return response.ok

    def associate_user_to_account(self, user_name, account_name, permission_set_name):
        """Associates an user with an account with proper permissions.

        Args:
            user_name: The name of the user to be assigned.
            account_name: Name of the account to which the user will be assigned
            permission_set_name: the Permission Set the user will have on the account

        Returns:
            bool: True or False

        """
        user = self.get_user_by_name(user_name)
        if not user:
            raise NoUser(user_name)
        account = self.get_account_by_name(account_name)
        if not account:
            raise NoAccount(account_name)
        profile_id = self._provision_application_profile_for_aws_account_instance(permission_set_name, account_name)
        if not profile_id:
            raise NoProfileID(f'{account_name}:{permission_set_name}')
        content_string = {'accessorId': user.id,
                          'accessorType': 'USER',
                          'accessorDisplay': {'userName': user_name,
                                              'firstName': user.first_name,
                                              'last_name': user.last_name,
                                              'windowsUpn': user_name},
                          'instanceId': account.instance_id,
                          'profileId': profile_id,
                          'directoryType': 'UserPool',
                          'directoryId': self.directory_id}
        payload = self.get_api_payload(content_string=content_string,
                                       target='AssociateProfile',
                                       path='/control/',
                                       x_amz_target='com.amazon.switchboard.service.SWBService.AssociateProfile',
                                       region=self.aws_region)
        self.logger.debug('Trying to assign users to aws account with payload: %s', payload)
        response = self.session.post(self.endpoint_url,
                                     json=payload)
        if not response.ok:
            self.logger.error('Received :%s', response.text)
        return response.ok

    def disassociate_user_from_account(self, user_name, account_name, permission_set_name):
        """Disassociates an user with an account with proper permissions.

        Args:
            user_name: The name of the user to be assigned.
            account_name: Name of the account to which the user will be assigned
            permission_set_name: the Permission Set the user will have on the account

        Returns:
            bool: True or False

        """
        user = self.get_user_by_name(user_name)
        if not user:
            raise NoUser(user_name)
        account = self.get_account_by_name(account_name)
        if not account:
            raise NoAccount(account_name)
        profile_id = self._get_aws_account_profile_for_permission_set(account_name,
                                                                      permission_set_name).get('profileId')
        if not profile_id:
            raise NoProfileID(f'{account_name}:{permission_set_name}')
        content_string = {'accessorId': user.id,
                          'accessorType': 'USER',
                          'accessorDisplay': {'userName': user_name,
                                              'firstName': user.first_name,
                                              'last_name': user.last_name,
                                              'windowsUpn': user_name},
                          'instanceId': account.instance_id,
                          'profileId': profile_id,
                          'directoryType': 'UserPool',
                          'directoryId': self.directory_id}
        payload = self.get_api_payload(content_string=content_string,
                                       target='DisassociateProfile',
                                       path='/control/',
                                       x_amz_target='com.amazon.switchboard.service.SWBService.DisassociateProfile',
                                       region=self.aws_region)
        self.logger.debug('Trying to disassociate users from aws account with payload: %s', payload)
        response = self.session.post(self.endpoint_url,
                                     json=payload)
        if not response.ok:
            self.logger.error('Received :%s', response.text)
        return response.ok

    def _get_paginated_results(self,  # pylint: disable=too-many-arguments, too-many-locals
                               content_payload,
                               path,
                               target,
                               amz_target,
                               object_group,
                               object_type=None,
                               region=None,
                               next_token_marker='NextToken',
                               url=None,
                               headers=None):
        payload = self.get_api_payload(content_string=content_payload,
                                       target=target,
                                       path=f'/{path}/' if path else '/',
                                       x_amz_target=amz_target,
                                       region=region)
        if headers:
            payload.update({'headers': headers})
        url = url or f'{self.api_url}/{path}'
        response, next_token = self._get_partial_response(url, payload, next_token_marker)
        for data in response.json().get(object_group, []):
            if object_type:
                yield object_type(self, data)
            else:
                yield data
        while next_token:
            content_string = copy.deepcopy(json.loads(payload.get('contentString')))
            content_string.update({next_token_marker: next_token})
            payload.update({'contentString': json.dumps(content_string)})
            response, next_token = self._get_partial_response(url, payload, next_token_marker)
            for data in response.json().get(object_group, []):
                if object_type:
                    yield object_type(self, data)
                else:
                    yield data

    def _get_partial_response(self, url, payload, next_token_marker):
        response = self.session.post(url, json=payload)
        if not response.ok:
            raise ValueError(response.text)
        next_token = response.json().get(next_token_marker)
        return response, next_token

    def create_permission_set(self, name, description=' ', relay_state=None, ttl='PT2H'):
        """Create a permission_set with a aws defined policy or custom policy.

        Args:
                    name: The name of the permission_set .
                    description: Description for the permission set
                    relay_state: The relay state for the permission set.
                                 https://docs.aws.amazon.com/singlesignon/latest/userguide/howtopermrelaystate.html
                    ttl: session duration
        Returns:
                    PermissionSet: Permission Set object

        """
        content_string = {'permissionSetName': name,
                          'description': description,
                          'relayState': relay_state or self.relay_state,
                          'ttl': ttl}
        payload = self.get_api_payload(content_string=content_string,
                                       target='CreatePermissionSet',
                                       path='/control/',
                                       x_amz_target='com.amazon.switchboard.service.SWBService.CreatePermissionSet',
                                       region=self.aws_region)
        self.logger.debug('Trying to create Permission set with payload: %s', payload)
        response = self.session.post(self.endpoint_url, json=payload)
        if response.ok:
            return PermissionSet(self, response.json().get('permissionSet'))
        return None

    def delete_permission_set(self,
                              name):
        """Delete a permission_set .

        Args:
                name: The name of the permission_set .

        Returns:
                Bool: Status of the deletion

        """
        permission_set_id = self.get_permission_set_by_name(name).id
        content_string = {'permissionSetId': permission_set_id}
        payload = self.get_api_payload(content_string=content_string,
                                       target='DeletePermissionSet',
                                       path='/control/',
                                       x_amz_target='com.amazon.switchboard.service.SWBService.DeletePermissionSet',
                                       region=self.aws_region
                                       )
        self.logger.debug('Trying to delete Permission set...')

        response = self.session.post(self.endpoint_url,
                                     json=payload)
        if not response.ok:
            self.logger.error('Received :%s', response.text)
        return response.ok



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File: controltower.py
#
# Copyright 2020 Costas Tyfoxylos
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to
#  deal in the Software without restriction, including without limitation the
#  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
#  sell copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
#  pylint: disable=too-many-lines

"""
Main code for controltower.

.. _Google Python Style Guide:
   http://google.github.io/styleguide/pyguide.html

"""

import copy
import itertools
import json
import logging
import time
from functools import wraps
from time import sleep

import boto3
import botocore
import requests

from opnieuw import retry
from awsapilib.authentication import Authenticator, LoggerMixin

from .controltowerexceptions import (UnsupportedTarget,
                                     OUCreating,
                                     NoServiceCatalogAccess,
                                     ServiceCallFailed,
                                     ControlTowerBusy,
                                     ControlTowerNotDeployed,
                                     PreDeployValidationFailed,
                                     EmailCheckFailed,
                                     EmailInUse,
                                     UnavailableRegion,
                                     RoleCreationFailure)
from .resources import (LOGGER,
                        LOGGER_BASENAME,
                        ServiceControlPolicy,
                        CoreAccount,
                        ControlTowerAccount,
                        ControlTowerOU,
                        AccountFactory,
                        OrganizationsOU,
                        GuardRail,
                        CREATING_ACCOUNT_ERROR_MESSAGE)

__author__ = '''Costas Tyfoxylos <ctyfoxylos@schubergphilis.com>'''
__docformat__ = '''google'''
__date__ = '''18-02-2020'''
__copyright__ = '''Copyright 2020, Costas Tyfoxylos'''
__credits__ = ["Costas Tyfoxylos"]
__license__ = '''MIT'''
__maintainer__ = '''Costas Tyfoxylos'''
__email__ = '''<ctyfoxylos@schubergphilis.com>'''
__status__ = '''Development'''  # "Prototype", "Development", "Production".


class ControlTower(LoggerMixin):  # pylint: disable=too-many-instance-attributes,too-many-public-methods
    """Models Control Tower by wrapping around service catalog."""

    api_content_type = 'application/x-amz-json-1.1'
    api_user_agent = 'aws-sdk-js/2.528.0 promise'
    supported_targets = ['listManagedOrganizationalUnits',
                         'manageOrganizationalUnit',
                         'deregisterOrganizationalUnit',
                         'listManagedAccounts',
                         'getGuardrailComplianceStatus',
                         'describeManagedOrganizationalUnit',
                         'listGuardrailsForTarget',
                         'getAvailableUpdates',
                         'describeCoreService',
                         'getAccountInfo',
                         'listEnabledGuardrails',
                         'listGuardrails',
                         'listOrganizationalUnitsForParent',
                         'listDriftDetails',
                         'getLandingZoneStatus',
                         'setupLandingZone',
                         'getHomeRegion',
                         'listGuardrailViolations',
                         'getCatastrophicDrift',
                         'getGuardrailComplianceStatus',
                         'describeAccountFactoryConfig',
                         'performPreLaunchChecks',
                         'deleteLandingZone',
                         ]
    core_account_types = ['PRIMARY', 'LOGGING', 'SECURITY']

    def validate_availability(method):  # noqa
        """Validation decorator."""

        @wraps(method)
        def wrap(*args, **kwargs):
            """Inner wrapper decorator."""
            logger = logging.getLogger(f'{LOGGER_BASENAME}.validation_decorator')
            control_tower_instance = args[0]
            logger.debug('Decorating method: %s', method)
            if not control_tower_instance.is_deployed:
                raise ControlTowerNotDeployed
            if control_tower_instance.busy:
                raise ControlTowerBusy
            return method(*args, **kwargs)  # pylint: disable=not-callable

        return wrap

    def __init__(self, arn, settling_time=90, region=None):
        self.aws_authenticator = Authenticator(arn, region=region)
        self.service_catalog = boto3.client('servicecatalog', **self.aws_authenticator.assumed_role_credentials)
        self.organizations = boto3.client('organizations', **self.aws_authenticator.assumed_role_credentials)
        self.session = self._get_authenticated_session()
        self._region = region
        self._is_deployed = None
        self.url = f'https://{self.region}.console.aws.amazon.com/controltower/api/controltower'
        self._iam_admin_url = f'https://{self.region}.console.aws.amazon.com/controltower/api/iamadmin'
        self._account_factory_ = None
        self.settling_time = settling_time
        self._root_ou = None
        self._core_accounts = None

    @property
    def _account_factory(self):
        if any([not self.is_deployed,
                self.percentage_complete != 100]):
            return None
        if self._account_factory_ is None:
            self._account_factory_ = self._get_account_factory(self.service_catalog)
        return self._account_factory_

    @property
    def is_deployed(self):
        """The deployment status of control tower."""
        if not self._is_deployed:
            caller_region = self.aws_authenticator.region
            url = f'https://{caller_region}.console.aws.amazon.com/controltower/api/controltower'
            payload = self._get_api_payload(content_string={},
                                            target='getLandingZoneStatus',
                                            region=caller_region)
            self.logger.debug('Trying to get the deployed status of the landing zone with payload "%s"', payload)
            response = self.session.post(url, json=payload)
            if not response.ok:
                self.logger.error('Failed to get the deployed status of the landing zone with response status '
                                  '"%s" and response text "%s"',
                                  response.status_code, response.text)
                raise ServiceCallFailed(payload)
            not_deployed_states = ('NOT_STARTED', 'DELETE_COMPLETED', 'DELETE_FAILED')
            self._is_deployed = response.json().get('LandingZoneStatus') not in not_deployed_states
        return self._is_deployed

    @property
    def region(self):
        """Region."""
        if not self.is_deployed:
            self._region = self.aws_authenticator.region
            return self._region
        if self._region is None:
            caller_region = self.aws_authenticator.region
            url = f'https://{caller_region}.console.aws.amazon.com/controltower/api/controltower'
            payload = self._get_api_payload(content_string={}, target='getHomeRegion', region=caller_region)
            response = self.session.post(url, json=payload)
            if not response.ok:
                raise ServiceCallFailed(payload)
            self._region = response.json().get('HomeRegion') or self.aws_authenticator.region
        return self._region

    @staticmethod
    def get_available_regions():
        """The regions that control tower can be active in.

        Returns:
            regions (list): A list of strings of the regions that control tower can be active in.

        """
        url = 'https://api.regional-table.region-services.aws.a2z.com/index.json'
        response = requests.get(url)
        if not response.ok:
            LOGGER.error('Failed to retrieve the info')
            return []
        return [entry.get('id', '').split(':')[1]
                for entry in response.json().get('prices')
                if entry.get('id').startswith('controltower')]

    @property
    @validate_availability
    def core_accounts(self):
        """The core accounts of the landing zone.

        Returns:
            core_accounts (list): A list of the primary, logging and security account.

        """
        if self._core_accounts is None:
            core_accounts = []
            for account_type in self.core_account_types:
                payload = self._get_api_payload(content_string={'AccountType': account_type},
                                                target='describeCoreService')
                response = self.session.post(self.url, json=payload)
                if not response.ok:
                    raise ServiceCallFailed(f'Service call failed with payload {payload}')
                core_accounts.append(CoreAccount(self, account_type, response.json()))
            self._core_accounts = core_accounts
        return self._core_accounts

    @property
    @validate_availability
    def root_ou(self):
        """The root ou of control tower.

        Returns:
            root_ou (ControlTowerOU): The root ou object.

        """
        if self._root_ou is None:
            self._root_ou = self.get_organizational_unit_by_name('Root')
        return self._root_ou

    def _get_authenticated_session(self):
        return self.aws_authenticator.get_control_tower_authenticated_session()

    @property
    def _active_artifact(self):
        artifacts = self.service_catalog.list_provisioning_artifacts(ProductId=self._account_factory.product_id)
        return next((artifact for artifact in artifacts.get('ProvisioningArtifactDetails', [])
                     if artifact.get('Active')),
                    None)

    @staticmethod
    def _get_account_factory(service_catalog_client):
        filter_ = {'Owner': ['AWS Control Tower']}
        try:
            return AccountFactory(service_catalog_client,
                                  service_catalog_client.search_products(Filters=filter_
                                                                         ).get('ProductViewSummaries', [''])[0])
        except IndexError:
            raise NoServiceCatalogAccess(('Please make sure the role used has access to the "AWS Control Tower Account '
                                          'Factory Portfolio" in Service Catalog under "Groups, roles, and users"'))

    def _validate_target(self, target):
        if target not in self.supported_targets:
            raise UnsupportedTarget(target)
        return target

    def _get_api_payload(self,  # pylint: disable=too-many-arguments
                         content_string,
                         target,
                         method='POST',
                         params=None,
                         path=None,
                         region=None):
        target = self._validate_target(target)
        payload = {'contentString': json.dumps(content_string),
                   'headers': {'Content-Type': self.api_content_type,
                               'X-Amz-Target': f'AWSBlackbeardService.{target[0].capitalize() + target[1:]}',
                               'X-Amz-User-Agent': self.api_user_agent},
                   'method': method,
                   'operation': target,
                   'params': params or {},
                   'path': path or '/',
                   'region': region or self.region}
        return copy.deepcopy(payload)

    def _get_paginated_results(self,  # pylint: disable=too-many-arguments
                               content_payload,
                               target,
                               object_group=None,
                               object_type=None,
                               method='POST',
                               params=None,
                               path=None,
                               region=None,
                               next_token_marker='NextToken'):
        payload = self._get_api_payload(content_string=content_payload,
                                        target=target,
                                        method=method,
                                        params=params,
                                        path=f'/{path}/' if path else '/',
                                        region=region)
        response, next_token = self._get_partial_response(payload, next_token_marker)
        if not object_group:
            yield response.json()
        else:
            for data in response.json().get(object_group, []):
                if object_type:
                    yield object_type(self, data)
                else:
                    yield data
        while next_token:
            content_string = copy.deepcopy(json.loads(payload.get('contentString')))
            content_string.update({next_token_marker: next_token})
            payload.update({'contentString': json.dumps(content_string)})
            response, next_token = self._get_partial_response(payload, next_token_marker)
            if not object_group:
                yield response.json()
            else:
                for data in response.json().get(object_group, []):
                    if object_type:
                        yield object_type(self, data)
                    else:
                        yield data

    def _get_partial_response(self, payload, next_token_marker):
        response = self.session.post(self.url, json=payload)
        if not response.ok:
            self.logger.debug('Failed getting partial response with payload :%s\n', payload)
            self.logger.debug('Response received :%s\n', response.content)
            raise ValueError(response.text)
        next_token = response.json().get(next_token_marker)
        return response, next_token

    @property
    def _update_data(self):
        return next(self._get_paginated_results(content_payload={},
                                                target='getAvailableUpdates'), {})

    @property
    @validate_availability
    def baseline_update_available(self):
        """Baseline update available."""
        return self._update_data.get('BaselineUpdateAvailable')

    @property
    @validate_availability
    def guardrail_update_available(self):
        """Guardrail update available."""
        return self._update_data.get('GuardrailUpdateAvailable')

    @property
    @validate_availability
    def landing_zone_update_available(self):
        """Landing Zone update available."""
        return self._update_data.get('LandingZoneUpdateAvailable')

    @property
    @validate_availability
    def service_landing_zone_version(self):
        """Service landing zone version."""
        return self._update_data.get('ServiceLandingZoneVersion')

    @property
    @validate_availability
    def user_landing_zone_version(self):
        """User landing zone version."""
        return self._update_data.get('UserLandingZoneVersion')

    @property
    @validate_availability
    def landing_zone_version(self):
        """Landing zone version."""
        return self._update_data.get('UserLandingZoneVersion')

    @property
    @validate_availability
    def organizational_units(self):
        """The organizational units under control tower.

        Returns:
            organizational_units (OrganizationalUnit): A list of organizational units objects under control tower's
            control.

        """
        return self._get_paginated_results(content_payload={'MaxResults': 20},
                                           target='listManagedOrganizationalUnits',
                                           object_type=ControlTowerOU,
                                           object_group='ManagedOrganizationalUnitList',
                                           next_token_marker='NextToken')

    @validate_availability
    def register_organizations_ou(self, name):
        """Registers an Organizations OU under control tower.

        Args:
            name (str): The name of the Organizations OU to register to Control Tower.

        Returns:
            result (bool): True if successfull, False otherwise.

        """
        if self.get_organizational_unit_by_name(name):
            self.logger.info('OU "%s" is already registered with Control Tower.', name)
            return True
        org_ou = self.get_organizations_ou_by_name(name)
        if not org_ou:
            self.logger.error('OU "%s" does not exist under organizations.', name)
            return False
        return self._register_org_ou_in_control_tower(org_ou)

    @validate_availability
    def create_organizational_unit(self, name):
        """Creates a Control Tower managed organizational unit.

        Args:
            name (str): The name of the OU to create.

        Returns:
            result (bool): True if successfull, False otherwise.

        """
        self.logger.debug('Trying to create OU :"%s" under root ou', name)
        try:
            response = self.organizations.create_organizational_unit(ParentId=self.root_ou.id, Name=name)
        except botocore.exceptions.ClientError as err:
            status = err.response["ResponseMetadata"]["HTTPStatusCode"]
            error_code = err.response["Error"]["Code"]
            error_message = err.response["Error"]["Message"]
            if not status == 200:
                self.logger.error('Failed to create OU "%s" under Organizations with error code %s: %s',
                                  name, error_code, error_message)
                return False
        org_ou = OrganizationsOU(response.get('OrganizationalUnit', {}))
        self.logger.debug(response)
        return self._register_org_ou_in_control_tower(org_ou)

    def _register_org_ou_in_control_tower(self, org_ou):
        self.logger.debug('Trying to move management of OU under Control Tower')
        payload = self._get_api_payload(content_string={'OrganizationalUnitId': org_ou.id,
                                                        'OrganizationalUnitName': org_ou.name},
                                        target='manageOrganizationalUnit')
        response = self.session.post(self.url, json=payload)
        if not response.ok:
            self.logger.error('Failed to register OU "%s" to Control Tower with response status "%s" '
                              'and response text "%s"',
                              org_ou.name, response.status_code, response.text)
            return False
        self.logger.debug('Successfully moved management of OU "%s" under Control Tower', org_ou.name)
        return True

    def _is_busy_with_ou_guardrails(self):
        """The status of guardrails application for OUs in control tower."""
        payload = self._get_api_payload(content_string={'OrganizationUnitStatus': 'IN_PROGRESS'},
                                        target='listManagedOrganizationalUnits')
        self.logger.debug('Trying to get the status of OU guardrails application with payload "%s"', payload)
        response = self.session.post(self.url, json=payload)
        if not response.ok:
            self.logger.error('Failed to get the status OU guardrails application with response status '
                              '"%s" and response text "%s"',
                              response.status_code, response.text)
            raise ServiceCallFailed(payload)
        return bool(response.json().get('ManagedOrganizationalUnitList'))

    @validate_availability
    def delete_organizational_unit(self, name):
        """Deletes a Control Tower managed organizational unit.

        Args:
            name (str): The name of the OU to delete.

        Returns:
            result (bool): True if successfull, False otherwise.

        """
        organizational_unit = self.get_organizational_unit_by_name(name)
        if not organizational_unit:
            self.logger.error('No organizational unit with name :"%s" registered with Control Tower', name)
            return False
        payload = self._get_api_payload(content_string={'OrganizationalUnitId': organizational_unit.id},
                                        target='deregisterOrganizationalUnit')
        self.logger.debug('Trying to unregister OU "%s" with payload "%s"', name, payload)
        response = self.session.post(self.url, json=payload)
        if not response.ok:
            self.logger.error('Failed to unregister OU "%s" with response status "%s" and response text "%s"',
                              name, response.status_code, response.text)
            return False
        self.logger.debug('Successfully unregistered management of OU "%s" from Control Tower', name)
        self.logger.debug('Trying to delete OU "%s" from Organizations', name)
        response = self.organizations.delete_organizational_unit(OrganizationalUnitId=organizational_unit.id)
        self.logger.debug(response)
        return bool(response.get('ResponseMetadata', {}).get('HTTPStatusCode') == 200)

    @validate_availability
    def get_organizational_unit_by_name(self, name):
        """Gets a Control Tower managed Organizational Unit by name.

        Args:
            name (str): The name of the organizational unit to retrieve.

        Returns:
            result (ControlTowerOU): A OU object on success, None otherwise.

        """
        return next((ou for ou in self.organizational_units if ou.name == name), None)

    @validate_availability
    def get_organizational_unit_by_id(self, id_):
        """Gets a Control Tower managed Organizational Unit by id.

        Args:
            id_ (str): The id of the organizational unit to retrieve.

        Returns:
            result (ControlTowerOU): A OU object on success, None otherwise.

        """
        return next((ou for ou in self.organizational_units if ou.id == id_), None)

    @property
    @validate_availability
    def organizations_ous(self):
        """The organizational units under Organizations.

        Returns:
            organizational_units (OrganizationsOU): A list of organizational units objects under Organizations.

        """
        paginator = self.organizations.get_paginator('list_organizational_units_for_parent')
        pages = paginator.paginate(ParentId=self.root_ou.id)
        response = list(itertools.chain.from_iterable([page['OrganizationalUnits'] for page in pages]))

        return [OrganizationsOU(data)
                for data in response]

    @validate_availability
    def get_organizations_ou_by_name(self, name):
        """Gets an Organizations managed Organizational Unit by name.

        Args:
            name (str): The name of the organizational unit to retrieve.

        Returns:
            result (OrganizationsOU): A OU object on success, None otherwise.

        """
        return next((ou for ou in self.organizations_ous if ou.name == name), None)

    @validate_availability
    def get_organizations_ou_by_id(self, id_):
        """Gets an Organizations managed Organizational Unit by id.

        Args:
            id_ (str): The id of the organizational unit to retrieve.

        Returns:
            result (OrganizationsOU): A OU object on success, None otherwise.

        """
        return next((ou for ou in self.organizations_ous if ou.id == id_), None)

    @validate_availability
    def get_organizations_ou_by_arn(self, arn):
        """Gets an Organizations managed Organizational Unit by arn.

        Args:
            arn (str): The arn of the organizational unit to retrieve.

        Returns:
            result (OrganizationsOU): A OU object on success, None otherwise.

        """
        return next((ou for ou in self.organizations_ous if ou.arn == arn), None)

    @property
    @validate_availability
    def accounts(self):
        """The accounts under control tower.

        Returns:
            accounts (Account): A list of account objects under control tower's control.

        """
        return self._get_paginated_results(content_payload={},
                                           target='listManagedAccounts',
                                           object_type=ControlTowerAccount,
                                           object_group='ManagedAccountList',
                                           next_token_marker='NextToken')

    @property
    def _service_catalog_accounts_data(self):
        products = self.service_catalog.search_provisioned_products()
        return [data for data in products.get('ProvisionedProducts', [])
                if data.get('Type', '') == 'CONTROL_TOWER_ACCOUNT']

    @validate_availability
    def get_available_accounts(self):
        """Retrieves the available accounts from control tower.

        Returns:
            accounts (Account): A list of available account objects under control tower's control.

        """
        return self._filter_for_status('AVAILABLE')

    @validate_availability
    def get_erroring_accounts(self):
        """Retrieves the erroring accounts from control tower.

        Returns:
            accounts (Account): A list of erroring account objects under control tower's control.

        """
        return self._filter_for_status('ERROR')

    @validate_availability
    def get_accounts_with_available_updates(self):
        """Retrieves the accounts that have available updates from control tower.

        Returns:
            accounts (Account): A list of account objects under control tower's control with available updates.

        """
        return [account for account in self.accounts if account.has_available_update]

    @validate_availability
    def get_updated_accounts(self):
        """Retrieves the accounts that have no available updates from control tower.

        Returns:
            accounts (Account): A list of account objects under control tower's control with no available updates.

        """
        return [account for account in self.accounts if not account.has_available_update]

    def get_changing_accounts(self):
        """Retrieves the under change accounts from control tower.

        Returns:
            accounts (Account): A list of under change account objects under control tower's control.

        """
        products = self.service_catalog.search_provisioned_products()

        return [ControlTowerAccount(self, {'AccountId': data.get('PhysicalId')})
                for data in products.get('ProvisionedProducts', [])
                if all([data.get('Type', '') == 'CONTROL_TOWER_ACCOUNT',
                        data.get('Status', '') == 'UNDER_CHANGE'])]

    def _filter_for_status(self, status):
        return [account for account in self.accounts if account.service_catalog_status == status]

    def _get_by_attribute(self, attribute, value):
        return next((account for account in self.accounts
                     if getattr(account, attribute) == value), None)

    def _get_service_catalog_data_by_account_id(self, account_id):
        return next((data for data in self._service_catalog_accounts_data
                     if data.get('PhysicalId') == account_id), None)

    @validate_availability
    def get_account_by_name(self, name):
        """Retrieves an account by name.

        Returns:
            account (Account): An account object that matches the name or None.

        """
        return self._get_by_attribute('name', name)

    @validate_availability
    def get_account_by_id(self, id_):
        """Retrieves an account by id.

        Returns:
            account (Account): An account object that matches the id or None.

        """
        return self._get_by_attribute('id', id_)

    @validate_availability
    def get_account_by_arn(self, arn):
        """Retrieves an account by arn.

        Returns:
            account (Account): An account object that matches the arn or None.

        """
        return self._get_by_attribute('arn', arn)

    @retry(retry_on_exceptions=OUCreating, max_calls_total=7, retry_window_after_first_call_in_seconds=60)
    @validate_availability
    def create_account(self,  # pylint: disable=too-many-arguments
                       account_name,
                       account_email,
                       organizational_unit,
                       product_name=None,
                       sso_first_name=None,
                       sso_last_name=None,
                       sso_user_email=None):
        """Creates a Control Tower managed account.

        Args:
            account_name (str): The name of the account.
            account_email (str): The email of the account.
            organizational_unit (str): The organizational unit that the account should be under.
            product_name (str): The product name, if nothing is provided it uses the account name.
            sso_first_name (str): The first name of the SSO user, defaults to "Control"
            sso_last_name (str): The last name of the SSO user, defaults to "Tower"
            sso_user_email (str): The email of the sso, if nothing is provided it uses the account email.

        Returns:
            result (bool): True on success, False otherwise.

        """
        product_name = product_name or account_name
        sso_user_email = sso_user_email or account_email
        sso_first_name = sso_first_name or 'Control'
        sso_last_name = sso_last_name or 'Tower'
        if not self.get_organizational_unit_by_name(organizational_unit):
            if not self.create_organizational_unit(organizational_unit):
                self.logger.error('Unable to create the organizational unit!')
                return False
        while self.busy:
            time.sleep(1)
        arguments = {'ProductId': self._account_factory.product_id,
                     'ProvisionedProductName': product_name,
                     'ProvisioningArtifactId': self._active_artifact.get('Id'),
                     'ProvisioningParameters': [{'Key': 'AccountName',
                                                 'Value': account_name},
                                                {'Key': 'AccountEmail',
                                                 'Value': account_email},
                                                {'Key': 'SSOUserFirstName',
                                                 'Value': sso_first_name},
                                                {'Key': 'SSOUserLastName',
                                                 'Value': sso_last_name},
                                                {'Key': 'SSOUserEmail',
                                                 'Value': sso_user_email},
                                                {'Key': 'ManagedOrganizationalUnit',
                                                 'Value': organizational_unit}]}
        try:
            response = self.service_catalog.provision_product(**arguments)
        except botocore.exceptions.ClientError as err:
            if CREATING_ACCOUNT_ERROR_MESSAGE in err.response['Error']['Message']:
                raise OUCreating
            raise
        response_metadata = response.get('ResponseMetadata', {})
        success = response_metadata.get('HTTPStatusCode') == 200
        if not success:
            self.logger.error('Failed to create account, response was :%s', response_metadata)
            return False
        # Making sure that eventual consistency is not a problem here,
        # we wait for control tower to be aware of the service catalog process
        while not self.busy:
            time.sleep(1)
        return True

    @property
    @validate_availability
    def service_control_policies(self):
        """The service control policies under organization.

        Returns:
            service_control_policies (list): A list of SCPs under the organization.

        """
        return [ServiceControlPolicy(data)
                for data in self.organizations.list_policies(Filter='SERVICE_CONTROL_POLICY').get('Policies', [])]

    @validate_availability
    def get_service_control_policy_by_name(self, name):
        """Retrieves a service control policy by name.

        Args:
            name (str): The name of the SCP to retrieve

        Returns:
            scp (ServiceControlPolicy): The scp if a match is found else None.

        """
        return next((scp for scp in self.service_control_policies
                     if scp.name == name), None)

    @validate_availability
    def update(self):
        """Updates the control tower to the next available version.

        Returns:
            bool: True on success, False on failure.

        """
        if not self.landing_zone_update_available:
            self.logger.warning('Landing zone does not seem to need update, is at version %s',
                                self.landing_zone_version)
            return False
        log_account = next((account for account in self.core_accounts if account.label == 'LOGGING'), None)
        if not log_account:
            raise ServiceCallFailed('Could not retrieve logging account to get the email.')
        security_account = next((account for account in self.core_accounts if account.label == 'SECURITY'), None)
        if not security_account:
            raise ServiceCallFailed('Could not retrieve security account to get the email.')
        payload = self._get_update_payload(log_account.email, security_account.email)
        self.logger.debug('Trying to update the landing zone with payload "%s"', payload)
        response = self.session.post(self.url, json=payload)
        if not response.ok:
            self.logger.error('Failed to update the landing zone with response status "%s" and response text "%s"',
                              response.status_code, response.text)
            return False
        self.logger.debug('Successfully started updating landing zone')
        # Making sure that eventual consistency is not a problem here,
        # we wait for control tower to be aware of the service catalog process
        while not self.busy:
            time.sleep(1)
        return True

    def _get_update_payload(self, log_account_email, security_account_email):
        """Updates the control tower up to 2.6 version.

        Returns:
            bool: True on success, False on failure.

        """
        content = {'HomeRegion': self.region,
                   'LogAccountEmail': log_account_email,
                   'SecurityAccountEmail': security_account_email}
        if self.landing_zone_version == '2.6':
            region_list = [{"Region": region,
                            "RegionConfigurationStatus": "ENABLED" if region in self.governed_regions else "DISABLED"}
                           for region in self.get_available_regions()]
            content.update({'SetupLandingZoneActionType': 'UPDATE',
                            'RegionConfigurationList': region_list})
        return self._get_api_payload(content_string=content,
                                     target='setupLandingZone')

    @property
    def busy(self):
        """Busy."""
        return any([self.status == 'IN_PROGRESS',
                    self.status == 'DELETE_IN_PROGRESS',
                    self.get_changing_accounts(),
                    self._is_busy_with_ou_guardrails()])

    @property
    def status(self):
        """Status."""
        return self._get_status().get('LandingZoneStatus')

    @property
    def percentage_complete(self):
        """Percentage complete."""
        return self._get_status().get('PercentageComplete')

    @property
    def deploying_messages(self):
        """Deploying messages."""
        return self._get_status().get('Messages')

    @property
    def region_metadata_list(self):
        """Region metadata list."""
        return self._get_status().get('RegionMetadataList')

    @property
    def governed_regions(self):
        """Governed regions."""
        return [region.get('Region')
                for region in self.region_metadata_list if region.get('RegionStatus') == 'GOVERNED']

    @property
    def not_governed_regions(self):
        """Not governed regions."""
        return [region.get('Region')
                for region in self.region_metadata_list if region.get('RegionStatus') == 'NOT_GOVERNED']

    def _get_status(self):
        payload = self._get_api_payload(content_string={},
                                        target='getLandingZoneStatus')
        self.logger.debug('Trying to get the landing zone status with payload "%s"', payload)
        response = self.session.post(self.url, json=payload)
        if not response.ok:
            self.logger.error('Failed to get the landing zone status with response status "%s" and response text "%s"',
                              response.status_code, response.text)
            return {}
        self.logger.debug('Successfully got landing zone status.')
        return response.json()

    @property
    @validate_availability
    def drift_messages(self):
        """Drift messages."""
        payload = self._get_api_payload(content_string={},
                                        target='listDriftDetails')
        self.logger.debug('Trying to get the drift messages of the landing zone with payload "%s"', payload)
        response = self.session.post(self.url, json=payload)
        if not response.ok:
            self.logger.error('Failed to get the drift message of the landing zone with response status "%s" and '
                              'response text "%s"',
                              response.status_code, response.text)
            return []
        return response.json().get('DriftDetails')

    @property
    @validate_availability
    def enabled_guard_rails(self):
        """Enabled guard rails."""
        output = []
        for result in self._get_paginated_results(content_payload={}, target='listEnabledGuardrails'):
            output.extend([GuardRail(self, data) for data in result.get('EnabledGuardrailList')])
        return output

    @property
    @validate_availability
    def guard_rails(self):
        """Guard rails."""
        output = []
        for result in self._get_paginated_results(content_payload={}, target='listGuardrails'):
            output.extend([GuardRail(self, data) for data in result.get('GuardrailList')])
        return output

    @property
    @validate_availability
    def guard_rails_violations(self):
        """List guard rails violations."""
        output = []
        for result in self._get_paginated_results(content_payload={}, target='listGuardrailViolations'):
            output.extend(result.get('GuardrailViolationList'))
        return output

    @property
    @validate_availability
    def catastrophic_drift(self):
        """List of catastrophic drift."""
        output = []
        for result in self._get_paginated_results(content_payload={}, target='getCatastrophicDrift'):
            output.extend(result.get('DriftDetails'))
        return output

    @property
    def _account_factory_config(self):
        """The config of the account factory."""
        payload = self._get_api_payload(content_string={},
                                        target='describeAccountFactoryConfig')
        self.logger.debug('Trying to get the account factory config of the landing zone with payload "%s"', payload)
        response = self.session.post(self.url, json=payload)
        if not response.ok:
            self.logger.error('Failed to get the the account factory config of the landing zone with response status '
                              '"%s" and response text "%s"',
                              response.status_code, response.text)
            return {}
        return response.json().get('AccountFactoryConfig')

    def _pre_deploy_check(self):
        """Pre deployment check."""
        payload = self._get_api_payload(content_string={},
                                        target='performPreLaunchChecks')
        self.logger.debug('Trying the pre deployment check with payload "%s"', payload)
        response = self.session.post(self.url, json=payload)
        if not response.ok:
            self.logger.error('Failed to do the pre deployment checks with response status '
                              '"%s" and response text "%s"',
                              response.status_code, response.text)
            return []
        return response.json().get('PreLaunchChecksResult')

    def is_email_used(self, email):
        """Check email for availability to be used or if it is already in use."""
        payload = self._get_api_payload(content_string={'AccountEmail': email},
                                        target='getAccountInfo')
        self.logger.debug('Trying to check email with payload "%s"', payload)
        response = self.session.post(self.url, json=payload)
        if not response.ok:
            self.logger.error('Failed to check for email with response status '
                              '"%s" and response text "%s"',
                              response.status_code, response.text)
            raise EmailCheckFailed(response.text)
        return response.json().get('AccountWithEmailExists')

    def _validate_regions(self, regions):
        available_regions = self.get_available_regions()
        if not set(available_regions).issuperset(set(regions)):
            raise UnavailableRegion(set(regions) - set(available_regions))
        return regions

    def _create_system_role(self, parameters):
        default_params = {'Action': 'CreateServiceRole',
                          'ContentType': 'JSON',
                          'ServicePrincipalName': 'controltower.amazonaws.com',
                          'TemplateVersion': 1}
        default_params.update(parameters)
        payload = {'headers': {'Content-Type': 'application/x-amz-json-1.1'},
                   'method': 'GET',
                   'params': default_params,
                   'path': '/',
                   'region': 'us-east-1'}
        self.logger.debug('Trying to system role with payload "%s"', payload)
        response = self.session.post(self._iam_admin_url, json=payload)
        try:
            if all([not response.ok,
                    response.status_code == 409,
                    response.json().get('Error', {}).get('Code') == 'EntityAlreadyExists]']):
                self.logger.error('Entity already exists, response status "%s" and response text "%s"',
                                  response.status_code, response.text)
                return True
        except ValueError:
            self.logger.error('Error on request, response status "%s" and response text "%s"',
                              response.status_code, response.text)
            return False
        if not response.ok:
            self.logger.error('Entity already exists, response status "%s" and response text "%s"',
                              response.status_code, response.text)
            return True
        self.logger.debug('Successfully created system role.')
        return True

    def _create_control_tower_admin(self):
        parameters = {'AmazonManagedPolicyArn': 'arn:aws:iam::aws:policy/service-role/AWSControlTowerServiceRolePolicy',
                      'Description': 'AWS Control Tower policy to manage AWS resources',
                      'PolicyName': 'AWSControlTowerAdminPolicy',
                      'RoleName': 'AWSControlTowerAdmin',
                      'TemplateName': 'AWSControlTowerAdmin',
                      'TemplateVersion': 2}
        return self._create_system_role(parameters)

    def _create_control_tower_cloud_trail_role(self):
        parameters = {'Description': 'AWS Cloud Trail assumes this role to create and '
                                     'publish Cloud Trail logs',
                      'PolicyName': 'AWSControlTowerCloudTrailRolePolicy',
                      'RoleName': 'AWSControlTowerCloudTrailRole',
                      'TemplateName': 'AWSControlTowerCloudTrailRole'}
        return self._create_system_role(parameters)

    def _create_control_tower_stack_set_role(self):
        parameters = {'Description': 'AWS CloudFormation assumes this role to deploy '
                                     'stacksets in accounts created by AWS Control Tower',
                      'PolicyName': 'AWSControlTowerStackSetRolePolicy',
                      'RoleName': 'AWSControlTowerStackSetRole',
                      'TemplateName': 'AWSControlTowerStackSetRole'}
        return self._create_system_role(parameters)

    def _create_control_tower_config_aggregator_role(self):
        parameters = {'AmazonManagedPolicyArn': 'arn:aws:iam::aws:policy/service-role/AWSConfigRoleForOrganizations',
                      'Description': 'AWS ControlTower needs this role to help in '
                                     'external config rule detection',
                      'RoleName': 'AWSControlTowerConfigAggregatorRoleForOrganizations',
                      'TemplateName': 'AWSControlTowerConfigAggregatorRole'}
        return self._create_system_role(parameters)

    def deploy(self,  # pylint: disable=too-many-arguments,too-many-locals
               logging_account_email,
               security_account_email,
               logging_account_name='Log Archive',
               security_account_name='Audit',
               core_ou_name='Security',
               custom_ou_name='Sandbox',
               regions=None,
               retries=10,
               wait=1):
        """Deploys control tower.

        Returns:
            bool: True on success, False on failure.

        """
        if self.is_deployed:
            self.logger.warning('Control tower does not seem to need deploying, already deployed.')
            return True
        regions = self._validate_regions(regions or [self.region])
        region_list = [{"Region": region, "RegionConfigurationStatus": "ENABLED" if region in regions else "DISABLED"}
                       for region in self.get_available_regions()]
        validation = self._pre_deploy_check()
        self.logger.debug('Got validation response %s.', validation)
        if not all([list(entry.values()).pop().get('Result') == 'SUCCESS' for entry in validation]):
            raise PreDeployValidationFailed(validation)
        invalid_emails = [email for email in [logging_account_email, security_account_email]
                          if self.is_email_used(email)]
        if invalid_emails:
            raise EmailInUse(invalid_emails)
        if not all([self._create_control_tower_admin(),
                    self._create_control_tower_cloud_trail_role(),
                    self._create_control_tower_stack_set_role(),
                    self._create_control_tower_config_aggregator_role()]):
            raise RoleCreationFailure('Unable to create required roles AWSControlTowerAdmin, '
                                      'AWSControlTowerCloudTrailRole, AWSControlTowerStackSetRole, '
                                      'AWSControlTowerConfigAggregatorRole, manual cleanup is required.')
        accounts = [{'Accounts': [{'AccountEmail': logging_account_email,
                                   'AccountName': logging_account_name,
                                   'AccountType': 'LOGGING'},
                                  {'AccountEmail': security_account_email,
                                   'AccountName': security_account_name,
                                   'AccountType': 'SECURITY'}],
                     'OrganizationalUnitName': core_ou_name,
                     'OrganizationalUnitType': 'CORE'},
                    {'OrganizationalUnitName': custom_ou_name,
                     'OrganizationalUnitType': 'CUSTOM'}]
        configuration = {'OrganizationStructure': accounts,
                         'RegionConfigurationList': region_list}
        payload = self._get_api_payload(content_string={'Configuration': configuration,
                                                        'HomeRegion': self.region,
                                                        'LogAccountEmail': logging_account_email,
                                                        'SecurityAccountEmail': security_account_email,
                                                        'RegionConfigurationList': region_list,
                                                        'SetupLandingZoneActionType': 'CREATE'},
                                        target='setupLandingZone')
        self.logger.debug('Trying to deploy control tower with payload "%s"', payload)
        return self._deploy(payload, retries, wait)

    def _deploy(self, payload, retries=10, wait=1):
        succeded = False
        while retries:
            response = self.session.post(self.url, json=payload)
            succeded = response.ok
            retries -= 1
            if response.ok:
                retries = 0
            if all([not response.ok,
                    retries]):
                self.logger.error('Failed to deploy control tower with response status "%s" and response text "%s"'
                                  'still have %s retries will wait for %s seconds', response.status_code,
                                  response.text, retries, wait)
                sleep(wait)
        if not succeded:
            self.logger.error('Failed to deploy control tower, retries were spent.. Maybe try again later?')
            return False
        self.logger.debug('Successfully started deploying control tower.')
        # Making sure that eventual consistency is not a problem here,
        # we wait for control tower to be aware of the service catalog process
        while not self.busy:
            time.sleep(1)
        return True

    def decommission(self):
        """Decommissions a landing zone.

        The api call does not seem to be enough and although the resources are decomissioned like with
        the proper process, control tower responds with a delete failed on the api, so it seems that
        aws needs to perform actions on their end for the decommissioning to be successful.

        Returns:
            response (bool): True if the process starts successfully, False otherwise.

        """
        payload = self._get_api_payload(content_string={},
                                        target='deleteLandingZone',
                                        region=self.region)
        response = self.session.post(self.url, json=payload)
        if not response.ok:
            self.logger.error('Failed to decommission control tower with response status "%s" and response text "%s"',
                              response.status_code, response.text)
            return False
        self._is_deployed = None
        self.logger.debug('Successfully started decommissioning control tower.')
        return True
